import logging

from stevedore import extension

from magnus import defaults
from magnus.executor import BaseExecutor

logger = logging.getLogger(defaults.LOGGER_NAME)
logging.getLogger("stevedore").setLevel(logging.CRITICAL)

# --8<-- [start:docs]


class BaseIntegration:
    """
    Base class for handling integration between Executor and one of Catalog, Secrets, RunLogStore.
    """

    executor_type = ""
    service_type = ""  # One of secret, catalog, datastore, experiment tracker
    service_provider = ""  # The actual implementation of the service

    def __init__(self, executor: "BaseExecutor", integration_service: object):
        self.executor = executor
        self.service = integration_service

    def validate(self, **kwargs):
        """
        Raise an exception if the executor_type is not compatible with service provider.

        By default, it is considered as compatible.
        """

    def configure_for_traversal(self, **kwargs):
        """
        Do any changes needed to both executor and service provider during traversal of the graph.

        By default, no change is required.
        """

    def configure_for_execution(self, **kwargs):
        """
        Do any changes needed to both executor and service provider during execution of a node.

        By default, no change is required.
        """


# --8<-- [end:docs]


def get_integration_handler(executor: "BaseExecutor", service: object) -> BaseIntegration:
    """
    Return the integration handler between executor and the service.

    If none found to be implemented, return the BaseIntegration which does nothing.

    Args:
        executor (BaseExecutor): The executor
        service (object): The service provider

    Returns:
        [BaseIntegration]: The implemented integration handler or BaseIntegration if none found

    Raises:
        Exception: If multiple integrations are found for the executor and service
    """
    service_type = service.service_type  # type: ignore
    service_name = getattr(service, "service_name")
    integrations = []

    # Get all the integrations defined by the 3rd party in their pyproject.toml
    mgr = extension.ExtensionManager(
        namespace="integration",
        invoke_on_load=True,
        invoke_kwds={"executor": executor, "integration_service": service},
    )
    for _, kls in mgr.items():
        if (
            kls.obj.executor_type == executor.service_name
            and kls.obj.service_type == service_type
            and kls.obj.service_provider == service_name
        ):
            logger.info(f"Identified an integration pattern {kls.obj}")
            integrations.append(kls.obj)

    # Get all the implementations defined by the magnus package
    for kls in BaseIntegration.__subclasses__():
        # Match the exact service type
        if kls.service_type == service_type and kls.service_provider == service_name:
            # Match either all executor or specific ones provided
            if kls.executor_type == "" or kls.executor_type == executor.service_name:
                integrations.append(kls(executor=executor, integration_service=service))

    if len(integrations) > 1:
        msg = (
            f"Multiple integrations between {executor.service_name} and {service_name} of type {service_type} found. "
            "If you defined an integration pattern, please ensure it is specific and does not conflict with magnus "
            " implementations."
        )
        logger.exception(msg)
        raise Exception(msg)

    if not integrations:
        logger.warning(
            f"Could not find an integration pattern for {executor.service_name} and {service_name} for {service_type}."
            " This implies that there is no need to change the configurations."
        )
        return BaseIntegration(executor, service)

    return integrations[0]


def validate(executor: "BaseExecutor", service: object, **kwargs):
    """
    Helper function to resolve the Integration class and validate the compatibility between executor and service

    Args:
        executor (BaseExecutor) : The executor
        service (object): The service provider
    """
    integration_handler = get_integration_handler(executor, service)
    integration_handler.validate(**kwargs)


def configure_for_traversal(executor: "BaseExecutor", service: object, **kwargs):
    """
    Helper function to resolve the Integration class and configure the executor and service for graph traversal

    Args:
        executor (BaseExecutor) : The executor
        service (object): The service provider
    """
    integration_handler = get_integration_handler(executor, service)
    integration_handler.configure_for_traversal(**kwargs)


def configure_for_execution(executor: "BaseExecutor", service: object, **kwargs):
    """
    Helper function to resolve the Integration class and configure the executor and service for execution

    Args:
        executor (BaseExecutor) : The executor
        service (object): The service provider
    """
    integration_handler = get_integration_handler(executor, service)
    integration_handler.configure_for_execution(**kwargs)


class BufferedRunLogStore(BaseIntegration):
    """
    Integration between any executor and buffered run log store
    """

    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "buffered"  # The actual implementation of the service

    def validate(self, **kwargs):
        if not self.executor.service_name == "local":
            raise Exception("Buffered run log store is only supported for local executor")

        msg = (
            "Run log generated by buffered run log store are not persisted. "
            "Re-running this run, in case of a failure, is not possible"
        )
        logger.warning(msg)


class DoNothingCatalog(BaseIntegration):
    """
    Integration between any executor and do nothing catalog
    """

    service_type = "catalog"  # One of secret, catalog, datastore
    service_provider = "do-nothing"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = "A do-nothing catalog does not hold any data and therefore cannot pass data between nodes."
        logger.warning(msg)


class DoNothingSecrets(BaseIntegration):
    """
    Integration between any executor and do nothing secrets
    """

    service_type = "secrets"  # One of secret, catalog, datastore
    service_provider = "do-nothing"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = "A do-nothing secrets does not hold any secrets and therefore cannot return you any secrets."
        logger.warning(msg)


class DoNothingExperimentTracker(BaseIntegration):
    """
    Integration between any executor and do nothing experiment tracker
    """

    service_type = "experiment_tracker"  # One of secret, catalog, datastore
    service_provider = "do-nothing"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = "A do-nothing experiment tracker does nothing and therefore cannot track anything."
        logger.warning(msg)



import logging

from magnus import defaults
from magnus.integration import BaseIntegration

logger = logging.getLogger(defaults.NAME)


class BufferedRunLogStore(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """

    executor_type = "k8s-job"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "buffered"  # The actual implementation of the service

    def validate(self, **kwargs):
        raise Exception("K8s job cannot run work with buffered run log store")


class FileSystemRunLogStore(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """

    executor_type = "k8s-job"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "file-system"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "K8s job cannot run work with file-system run log store."
            "Unless you have made a mechanism to use volume mounts"
        )
        logger.warning(msg)


class ChunkedFSRunLogStore(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """

    executor_type = "k8s-job"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "chunked-fs"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "K8s job cannot run work with chunked-fs run log store."
            "Unless you have made a mechanism to use volume mounts"
        )
        logger.warning(msg)


class FileSystemCatalog(BaseIntegration):
    """
    Only local execution mode is possible for Buffered Run Log store
    """

    executor_type = "k8s-job"
    service_type = "catalog"  # One of secret, catalog, datastore
    service_provider = "file-system"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = (
            "K8s Job cannot run work with file-system catalog." "Unless you have made a mechanism to use volume mounts"
        )
        logger.warning(msg)



from __future__ import annotations

import logging
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, computed_field, field_validator, model_validator
from rich import print
from ruamel.yaml import YAML
from typing_extensions import Self

from magnus import defaults, entrypoints, graph, utils
from magnus.extensions.nodes import FailNode, MapNode, ParallelNode, StubNode, SuccessNode, TaskNode
from magnus.nodes import TraversalNode

logger = logging.getLogger(defaults.LOGGER_NAME)

StepType = Union["Stub", "Task", "Success", "Fail", "Parallel", "Map"]
TraversalTypes = Union["Stub", "Task", "Parallel", "Map"]


ALLOWED_COMMAND_TYPES = ["shell", "python", "notebook"]


class Catalog(BaseModel):
    """
    Use to instruct a task to sync data from/to the central catalog.
    Please refer to [concepts](concepts/catalog.md) for more information.

    Attributes:
        get (List[str]): List of glob patterns to get from central catalog to the compute data folder.
        put (List[str]): List of glob patterns to put into central catalog from the compute data folder.

    Examples:
        >>> from magnus import Catalog, Task
        >>> catalog = Catalog(compute_data_folder="/path/to/data", get=["*.csv"], put=["*.csv"])

        >>> task = Task(name="task", catalog=catalog, command="echo 'hello'")

    """

    model_config = ConfigDict(extra="forbid")  # Need to be for command, would be validated later
    # Note: compute_data_folder was confusing to explain, might be introduced later.
    # compute_data_folder: str = Field(default="", alias="compute_data_folder")
    get: List[str] = Field(default_factory=list, alias="get")
    put: List[str] = Field(default_factory=list, alias="put")


class BaseTraversal(ABC, BaseModel):
    name: str
    next_node: str = Field(default="", alias="next")
    terminate_with_success: bool = Field(default=False, exclude=True)
    terminate_with_failure: bool = Field(default=False, exclude=True)
    on_failure: str = Field(default="", alias="on_failure")

    model_config = ConfigDict(extra="forbid")

    @computed_field  # type: ignore
    @property
    def internal_name(self) -> str:
        return self.name

    def __rshift__(self, other: StepType) -> StepType:
        if self.next_node:
            raise Exception(f"The node {self} already has a next node: {self.next_node}")
        self.next_node = other.name

        return other

    def __lshift__(self, other: TraversalNode) -> TraversalNode:
        if other.next_node:
            raise Exception(f"The {other} node already has a next node: {other.next_node}")
        other.next_node = self.name

        return other

    def depends_on(self, node: StepType) -> Self:
        assert not isinstance(node, Success)
        assert not isinstance(node, Fail)

        if node.next_node:
            raise Exception(f"The {node} node already has a next node: {node.next_node}")

        node.next_node = self.name
        return self

    @model_validator(mode="after")
    def validate_terminations(self) -> Self:
        if self.terminate_with_failure and self.terminate_with_success:
            raise AssertionError("A node cannot terminate with success and failure")

        if self.terminate_with_failure or self.terminate_with_success:
            if self.next_node and self.next_node not in ["success", "fail"]:
                raise AssertionError("A node being terminated cannot have a user defined next node")

        if self.terminate_with_failure:
            self.next_node = "fail"

        if self.terminate_with_success:
            self.next_node = "success"

        return self

    @abstractmethod
    def create_node(self) -> TraversalNode:
        ...


class Task(BaseTraversal):
    """
    An execution node of the pipeline.
    Please refer to [concepts](concepts/task.md) for more information.

    Attributes:
        name (str): The name of the node.
        command (str): The command to execute.

            - For python functions, [dotted path](concepts/task.md/#python_functions) to the function.
            - For shell commands: command to execute in the shell.
            - For notebooks: path to the notebook.
        command_type (str): The type of command to execute.
            Can be one of "shell", "python", or "notebook".
        catalog (Optional[Catalog]): The catalog to sync data from/to.
            Please see Catalog about the structure of the catalog.
        overrides (Dict[str, Any]): Any overrides to the command.
            Individual tasks can override the global configuration config by referring to the
            specific override.

            For example,
            ### Global configuration
            ```yaml
            executor:
              type: local-container
              config:
                docker_image: "magnus/magnus:latest"
                overrides:
                  custom_docker_image:
                    docker_image: "magnus/magnus:custom"
            ```
            ### Task specific configuration
            ```python
            task = Task(name="task", command="echo 'hello'", command_type="shell",
                    overrides={'local-container': custom_docker_image})
            ```
        notebook_output_path (Optional[str]): The path to save the notebook output.
            Only used when command_type is 'notebook', defaults to command+_out.ipynb
        optional_ploomber_args (Optional[Dict[str, Any]]): Any optional ploomber args.
            Only used when command_type is 'notebook', defaults to {}
        output_cell_tag (Optional[str]): The tag of the output cell.
            Only used when command_type is 'notebook', defaults to "magnus_output"
        terminate_with_failure (bool): Whether to terminate the pipeline with a failure after this node.
        terminate_with_success (bool): Whether to terminate the pipeline with a success after this node.
        on_failure (str): The name of the node to execute if the step fails.

    """

    command: str = Field(alias="command")
    command_type: str = Field(default="python")
    catalog: Optional[Catalog] = Field(default=None, alias="catalog")
    overrides: Dict[str, Any] = Field(default_factory=dict, alias="overrides")

    notebook_output_path: Optional[str] = Field(default=None, alias="notebook_output_path")
    optional_ploomber_args: Optional[Dict[str, Any]] = Field(default=None, alias="optional_ploomber_args")
    output_cell_tag: Optional[str] = Field(default=None, alias="output_cell_tag")

    @field_validator("command_type", mode="before")
    @classmethod
    def validate_command_type(cls, value: str) -> str:
        if value not in ALLOWED_COMMAND_TYPES:
            raise ValueError(f"Invalid command_type: {value}")
        return value

    @model_validator(mode="after")
    def check_notebook_args(self) -> "Task":
        if self.command_type != "notebook":
            assert (
                self.notebook_output_path is None
            ), "Only command_types of 'notebook' can be used with notebook_output_path"

            assert (
                self.optional_ploomber_args is None
            ), "Only command_types of 'notebook' can be used with optional_ploomber_args"

            assert self.output_cell_tag is None, "Only command_types of 'notebook' can be used with output_cell_tag"
        return self

    def create_node(self) -> TaskNode:
        if not self.next_node:
            if not (self.terminate_with_failure or self.terminate_with_success):
                raise AssertionError("A node not being terminated must have a user defined next node")
        return TaskNode.parse_from_config(self.model_dump(exclude_none=True))


class Stub(BaseTraversal):
    """
    A node that does nothing.

    A stub node can tak arbitrary number of arguments.
    Please refer to [concepts](concepts/stub.md) for more information.

    Attributes:
        name (str): The name of the node.
        terminate_with_failure (bool): Whether to terminate the pipeline with a failure after this node.
        terminate_with_success (bool): Whether to terminate the pipeline with a success after this node.

    """

    model_config = ConfigDict(extra="allow")
    catalog: Optional[Catalog] = Field(default=None, alias="catalog")

    def create_node(self) -> StubNode:
        if not self.next_node:
            if not (self.terminate_with_failure or self.terminate_with_success):
                raise AssertionError("A node not being terminated must have a user defined next node")

        return StubNode.parse_from_config(self.model_dump(exclude_none=True))


class Parallel(BaseTraversal):
    """
    A node that executes multiple branches in parallel.
    Please refer to [concepts](concepts/parallel.md) for more information.

    Attributes:
        name (str): The name of the node.
        branches (Dict[str, Pipeline]): A dictionary of branches to execute in parallel.
        terminate_with_failure (bool): Whether to terminate the pipeline with a failure after this node.
        terminate_with_success (bool): Whether to terminate the pipeline with a success after this node.
        on_failure (str): The name of the node to execute if any of the branches fail.
    """

    branches: Dict[str, "Pipeline"]

    @computed_field  # type: ignore
    @property
    def graph_branches(self) -> Dict[str, graph.Graph]:
        return {name: pipeline._dag.model_copy() for name, pipeline in self.branches.items()}

    def create_node(self) -> ParallelNode:
        if not self.next_node:
            if not (self.terminate_with_failure or self.terminate_with_success):
                raise AssertionError("A node not being terminated must have a user defined next node")

        node = ParallelNode(name=self.name, branches=self.graph_branches, internal_name="", next_node=self.next_node)
        return node


class Map(BaseTraversal):
    """
    A node that iterates over a list of items and executes a pipeline for each item.
    Please refer to [concepts](concepts/map.md) for more information.

    Attributes:
        branch: The pipeline to execute for each item.

        iterate_on: The name of the parameter to iterate over.
            The parameter should be defined either by previous steps or statically at the start of execution.

        iterate_as: The name of the iterable to be passed to functions.


        overrides (Dict[str, Any]): Any overrides to the command.

    """

    branch: "Pipeline"
    iterate_on: str
    iterate_as: str
    overrides: Dict[str, Any] = Field(default_factory=dict)

    @computed_field  # type: ignore
    @property
    def graph_branch(self) -> graph.Graph:
        return self.branch._dag.model_copy()

    def create_node(self) -> MapNode:
        if not self.next_node:
            if not (self.terminate_with_failure or self.terminate_with_success):
                raise AssertionError("A node not being terminated must have a user defined next node")

        node = MapNode(
            name=self.name,
            branch=self.graph_branch,
            internal_name="",
            next_node=self.next_node,
            iterate_on=self.iterate_on,
            iterate_as=self.iterate_as,
            overrides=self.overrides,
        )

        return node


class Success(BaseModel):
    """
    A node that represents a successful execution of the pipeline.

    Most often, there is no need to use this node as nodes can be instructed to
    terminate_with_success and pipeline with add_terminal_nodes=True.

    Attributes:
        name (str): The name of the node.
    """

    name: str = "success"

    @computed_field  # type: ignore
    @property
    def internal_name(self) -> str:
        return self.name

    def create_node(self) -> SuccessNode:
        return SuccessNode.parse_from_config(self.model_dump())


class Fail(BaseModel):
    """
    A node that represents a failed execution of the pipeline.

    Most often, there is no need to use this node as nodes can be instructed to
    terminate_with_failure and pipeline with add_terminal_nodes=True.

    Attributes:
        name (str): The name of the node.
    """

    name: str = "fail"

    @computed_field  # type: ignore
    @property
    def internal_name(self) -> str:
        return self.name

    def create_node(self) -> FailNode:
        return FailNode.parse_from_config(self.model_dump())


class Pipeline(BaseModel):
    """
    A Pipeline is a directed acyclic graph of Steps that define a workflow.

    Attributes:
        steps (List[Stub | Task | Parallel | Map | Success | Fail]): A list of Steps that make up the Pipeline.
        start_at (Stub | Task | Parallel | Map): The name of the first Step in the Pipeline.
        name (str, optional): The name of the Pipeline. Defaults to "".
        description (str, optional): A description of the Pipeline. Defaults to "".
        add_terminal_nodes (bool, optional): Whether to add terminal nodes to the Pipeline. Defaults to True.

    The default behavior is to add "success" and "fail" nodes to the Pipeline.
    To add custom success and fail nodes, set add_terminal_nodes=False and create success
    and fail nodes manually.

    """

    steps: List[StepType]
    start_at: TraversalTypes
    name: str = ""
    description: str = ""
    add_terminal_nodes: bool = True  # Adds "success" and "fail" nodes

    internal_branch_name: str = ""

    _dag: graph.Graph = PrivateAttr()
    model_config = ConfigDict(extra="forbid")

    def model_post_init(self, __context: Any) -> None:
        self.steps = [model.model_copy(deep=True) for model in self.steps]

        self._dag = graph.Graph(
            start_at=self.start_at.name,
            description=self.description,
            internal_branch_name=self.internal_branch_name,
        )

        for step in self.steps:
            if step.name == self.start_at.name:
                if isinstance(step, Success) or isinstance(step, Fail):
                    raise Exception("A success or fail node cannot be the start_at of the graph")
                assert step.next_node
            self._dag.add_node(step.create_node())

        if self.add_terminal_nodes:
            self._dag.add_terminal_nodes()

        self._dag.check_graph()

    def execute(
        self,
        configuration_file: str = "",
        run_id: str = "",
        tag: str = "",
        parameters_file: str = "",
        use_cached: str = "",
        log_level: str = defaults.LOG_LEVEL,
        output_pipeline_definition: str = "magnus-pipeline.yaml",
    ):
        """
        *Execute* the Pipeline.

        Execution of pipeline could either be:

        Traverse and execute all the steps of the pipeline, eg. [local execution](configurations/executors/local.md).

        Or create the ```yaml``` representation of the pipeline for other executors.

        Please refer to [concepts](concepts/executor.md) for more information.

        Args:
            configuration_file (str, optional): The path to the configuration file. Defaults to "".
                The configuration file can be overridden by the environment variable MAGNUS_CONFIGURATION_FILE.

            run_id (str, optional): The ID of the run. Defaults to "".
            tag (str, optional): The tag of the run. Defaults to "".
                Use to group multiple runs.

            parameters_file (str, optional): The path to the parameters file. Defaults to "".
            use_cached (str, optional): Whether to use cached results. Defaults to "".
                Provide the run_id of the older execution to recover.

            log_level (str, optional): The log level. Defaults to defaults.LOG_LEVEL.
            output_pipeline_definition (str, optional): The path to the output pipeline definition file.
                Defaults to "magnus-pipeline.yaml".

                Only applicable for the execution via SDK for non ```local``` executors.
        """
        from magnus.extensions.executor.local.implementation import LocalExecutor
        from magnus.extensions.executor.mocked.implementation import MockedExecutor

        logger.setLevel(log_level)

        run_id = utils.generate_run_id(run_id=run_id)
        configuration_file = os.environ.get("MAGNUS_CONFIGURATION_FILE", configuration_file)
        run_context = entrypoints.prepare_configurations(
            configuration_file=configuration_file,
            run_id=run_id,
            tag=tag,
            parameters_file=parameters_file,
            use_cached=use_cached,
        )

        run_context.execution_plan = defaults.EXECUTION_PLAN.CHAINED.value
        utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

        dag_definition = self._dag.model_dump(by_alias=True, exclude_none=True)

        run_context.dag = graph.create_graph(dag_definition)

        print("Working with context:")
        print(run_context)

        if not (isinstance(run_context.executor, LocalExecutor) or isinstance(run_context.executor, MockedExecutor)):
            logger.debug(run_context.dag.model_dump(by_alias=True))
            yaml = YAML()

            with open(output_pipeline_definition, "w", encoding="utf-8") as f:
                yaml.dump(
                    {"dag": run_context.dag.model_dump(by_alias=True, exclude_none=True)},
                    f,
                )

            return

        # Prepare for graph execution
        run_context.executor.prepare_for_graph_execution()

        logger.info("Executing the graph")
        run_context.executor.execute_graph(dag=run_context.dag)

        return run_context.run_log_store.get_run_log_by_id(run_id=run_context.run_id)



from typing import List, Optional

from magnus.datastore import DataCatalog


def is_catalog_out_of_sync(catalog, synced_catalogs=Optional[List[DataCatalog]]) -> bool:
    """
    Check if the catalog items are out of sync from already cataloged objects.
    If they are, return False.
    If the object does not exist or synced catalog does not exist, return True
    """
    if not synced_catalogs:
        return True  # If nothing has been synced in the past

    for synced_catalog in synced_catalogs:
        if synced_catalog.catalog_relative_path == catalog.catalog_relative_path:
            if synced_catalog.data_hash == catalog.data_hash:
                return False
            return True

    return True  # The object does not exist, sync it



from __future__ import annotations

import hashlib
import json
import logging
import os
import subprocess
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from string import Template as str_template
from typing import TYPE_CHECKING, Any, Dict, Mapping, Tuple, Union

from ruamel.yaml import YAML
from stevedore import driver

import magnus.context as context
from magnus import defaults, names
from magnus.defaults import TypeMapVariable

if TYPE_CHECKING:  # pragma: no cover
    from magnus.extensions.nodes import TaskNode
    from magnus.nodes import BaseNode


logger = logging.getLogger(defaults.LOGGER_NAME)
logging.getLogger("stevedore").setLevel(logging.CRITICAL)


def does_file_exist(file_path: str) -> bool:
    """Check if a file exists.
    Implemented here to avoid repetition of logic.

    Args:
        file_path (str): The file path to check

    Returns:
        bool: False if it does not otherwise True
    """
    my_file = Path(file_path)
    return my_file.is_file()


def does_dir_exist(file_path: Union[str, Path]) -> bool:
    """Check if a directory exists.
    Implemented here to avoid repetition of logic.

    Args:
        file_path (str or Path): The directory path to check

    Returns:
        bool: False if the directory does not exist, True otherwise
    """
    my_file = Path(file_path)
    return my_file.is_dir()


def safe_make_dir(directory: Union[str, Path]):
    """Safely make the directory.
    Ignore if it exists and create the parents if necessary.

    Args:
        directory (str): The directory path to create
    """
    Path(directory).mkdir(parents=True, exist_ok=True)


def generate_run_id(run_id: str = "") -> str:
    """Generate a new run_id.

    If the input run_id is none, we create one based on time stamp.

    Args:
        run_id (str, optional): Input Run ID. Defaults to None

    Returns:
        str: A generated run_id
    """
    # If we are not provided with a run_id, generate one
    if not run_id:
        now = datetime.now()
        run_id = f"{names.get_random_name()}-{now.hour:02}{now.minute:02}"

    return run_id


def apply_variables(apply_to: Dict[str, Any], variables: Dict[str, str]) -> Dict[str, Any]:
    """Safely applies the variables to a config.

    For example: For config:
         {'a' : ${b}}, the value of ${b} is replaced by b in the variables.

    If the ${b} does not exist in the variables, it is ignored in the config.

    Args:
        apply_to (dict): The config to apply variables
        variables (dict): The variables in key, value pairs

    Raises:
        Exception: If the variables is not dict

    Returns:
        dict: A transformed dict with variables applied
    """
    if not isinstance(variables, dict):
        raise Exception("Argument Variables should be dict")

    json_d = json.dumps(apply_to)
    transformed = str_template(json_d).substitute(**variables)
    return json.loads(transformed)


def get_module_and_attr_names(command: str) -> Tuple[str, str]:
    """Given a string of module.function, this functions returns the module name and func names.

    It also checks to make sure that the string is of expected 'module.func' format

    Args:
        command (str): String of format module.function_name

    Raises:
        Exception: If the string is of not format

    Returns:
        Tuple[str, str]: (module_name, function_name) extracted from the input string
    """
    mods = command.split(".")
    if len(mods) <= 1:
        raise Exception("The command should be a function to call")
    func = mods[-1]
    module = ".".join(mods[:-1])
    return module, func


def get_dag_hash(dag: Dict[str, Any]) -> str:
    """Generates the hash of the dag definition.

    Args:
        dag (dict): The dictionary object containing the dag definition

    Returns:
        str: The hash of the dag definition
    """
    dag_str = json.dumps(dag, sort_keys=True, ensure_ascii=True)
    return hashlib.sha1(dag_str.encode("utf-8")).hexdigest()


def load_yaml(file_path: str, load_type: str = "safe") -> Dict[str, Any]:
    """Loads an yaml and returns the dictionary.

    Args:
        file_path (str): The path of the yamlfile
        load_type (str, optional): The load type as understood by ruamel. Defaults to 'safe'.

    Returns:
        dict: The mapping as defined in the yaml file
    """
    with open(file_path, encoding="utf-8") as f:
        yaml = YAML(typ=load_type, pure=True)
        yaml_config = yaml.load(f)
    return yaml_config


def is_a_git_repo() -> bool:
    """Does a git command to see if the project is git versioned.

    Returns:
        bool: True if it is git versioned, False otherwise
    """
    command = "git rev-parse --is-inside-work-tree"
    try:
        subprocess.check_output(command.split()).strip().decode("utf-8")
        logger.info("Found the code to be git versioned")
        return True
    except BaseException:  # pylint: disable=W0702
        logger.error("No git repo found, unsafe hash")

    return False


def get_current_code_commit() -> Union[str, None]:
    """Gets the git sha id if the project is version controlled.

    Returns:
        Union[str, None]: SHA ID if the code is versioned, None otherwise
    """
    command = "git rev-parse HEAD"
    if not is_a_git_repo():
        return None
    try:
        label = subprocess.check_output(command.split()).strip().decode("utf-8")
        logger.info("Found the git commit to be: %s", label)
        return label
    except BaseException:  # pylint: disable=W0702
        logger.exception("Error getting git hash")
        raise


def archive_git_tracked(name: str):
    """Generate a git archive of the tracked files.

    Args:
        name (str): The name to give the archive

    Raises:
        Exception: If its not a git repo
    """
    command = f"git archive -v -o {name}.tar.gz --format=tar.gz HEAD"

    if not is_a_git_repo():
        raise Exception("Not a git repo")
    try:
        subprocess.check_output(command.split()).strip().decode("utf-8")
    except BaseException:  # pylint: disable=W0702
        logger.exception("Error archiving repo")
        raise


def is_git_clean() -> Tuple[bool, Union[None, str]]:
    """Checks if the git tree is clean and there are no modified tracked files.

    Returns:
        Tuple[bool, Union[None, str]]: None if its clean, comma-seperated file names if it is changed
    """
    command = "git diff --name-only"
    if not is_a_git_repo():
        return False, None
    try:
        label = subprocess.check_output(command.split()).strip().decode("utf-8")
        if not label:
            return True, None
        return False, label
    except BaseException:  # pylint: disable=W0702
        logger.exception("Error checking if the code is git clean")

    return False, None


def get_git_remote() -> Union[str, None]:
    """Gets the remote URL of git.

    Returns:
        Union[str, None]: Remote URL if the code is version controlled, None otherwise
    """
    command = "git config --get remote.origin.url"
    if not is_a_git_repo():
        return None
    try:
        label = subprocess.check_output(command.split()).strip().decode("utf-8")
        logger.info("Found the git remote to be: %s", label)
        return label
    except BaseException:  # pylint: disable=W0702
        logger.exception("Error getting git remote")
        raise


def get_local_docker_image_id(image_name: str) -> str:
    """If we are running in local settings, return the docker image id.

    Args:
        image_name (str): The image name we need the digest for

    Returns:
        str: The docker image digest
    """
    try:
        import docker

        client = docker.from_env()
        image = client.images.get(image_name)
        return image.attrs["Id"]
    except ImportError:  # pragma: no cover
        logger.warning("Did not find docker installed, some functionality might be affected")
    except BaseException:
        logger.exception(f"Could not find the image by name {image_name}")

    return ""


def get_git_code_identity():
    """Returns a code identity object for version controlled code.

    Args:
        run_log_store (magnus.datastore.BaseRunLogStore): The run log store used in this process

    Returns:
        magnus.datastore.CodeIdentity: The code identity used by the run log store.
    """
    code_identity = context.run_context.run_log_store.create_code_identity()
    try:
        code_identity.code_identifier = get_current_code_commit()
        code_identity.code_identifier_type = "git"
        code_identity.code_identifier_dependable, changed = is_git_clean()
        code_identity.code_identifier_url = get_git_remote()
        if changed:
            code_identity.code_identifier_message = "changes found in " + ", ".join(changed.split("\n"))
    except BaseException:
        logger.exception("Git code versioning problems")

    return code_identity


def remove_prefix(text: str, prefix: str) -> str:
    """Removes a prefix if one is present in the input text.

    Args:
        text (str): The input text to remove the prefix from
        prefix (str): The prefix that has to be removed

    Returns:
        str: The original string if no prefix is found, or the right prefix chomped string if present
    """
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text  # or whatever is given


def get_tracked_data() -> Dict[str, str]:
    """Scans the environment variables to find any user tracked variables that have a prefix MAGNUS_TRACK_
    Removes the environment variable to prevent any clashes in the future steps.

    Returns:
        dict: A dictionary of user tracked data
    """
    tracked_data = {}
    for env_var, value in os.environ.items():
        if env_var.startswith(defaults.TRACK_PREFIX):
            key = remove_prefix(env_var, defaults.TRACK_PREFIX)
            try:
                tracked_data[key.lower()] = json.loads(value)
            except json.decoder.JSONDecodeError:
                logger.warning(f"Tracker {key} could not be JSON decoded, adding the literal value")
                tracked_data[key.lower()] = value

            del os.environ[env_var]
    return tracked_data


def diff_dict(d1: Dict[str, Any], d2: Dict[str, Any]) -> Dict[str, Any]:
    """
    Given two dicts d1 and d2, return a new dict that has upsert items from d1.

    Args:
        d1 (reference): The reference dict.
        d2 (compare): Any new or modified items compared to d1 would be returned back

    Returns:
        dict: Any new or modified items in d2 in comparison to d1 would be sent back
    """
    diff = {}

    for k2, v2 in d2.items():
        if k2 in d1 and d1[k2] != v2:
            diff[k2] = v2
            continue
        diff[k2] = v2

    return diff


def hash_bytestr_iter(bytesiter, hasher, ashexstr=True):  # pylint: disable=C0116
    """Hashes the given bytesiter using the given hasher."""
    for block in bytesiter:  # pragma: no cover
        hasher.update(block)
    return hasher.hexdigest() if ashexstr else hasher.digest()  # pragma: no cover


def file_as_blockiter(afile, blocksize=65536):  # pylint: disable=C0116
    """From a StackOverflow answer: that is used to generate a MD5 hash of a large files.
    # https://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file.

    """
    with afile:  # pragma: no cover
        block = afile.read(blocksize)
        while len(block) > 0:
            yield block
            block = afile.read(blocksize)


def get_data_hash(file_name: str):
    """Returns the hash of the data file.

    Args:
        file_name (str): The file name to generated the hash

    Returns:
        str: The SHA ID of the file contents
    """
    # https://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file
    return hash_bytestr_iter(file_as_blockiter(open(file_name, "rb")), hashlib.sha256())  # pragma: no cover


def get_node_execution_command(
    node: BaseNode,
    map_variable: TypeMapVariable = None,
    over_write_run_id: str = "",
) -> str:
    """A utility function to standardize execution call to a node via command line.

    Args:
        executor (object): The executor class.
        node (object): The Node to execute
        map_variable (str, optional): If the node belongs to a map step. Defaults to None.

    Returns:
        str: The execution command to run a node via command line.
    """
    run_id = context.run_context.run_id

    if over_write_run_id:
        run_id = over_write_run_id

    log_level = logging.getLevelName(logger.getEffectiveLevel())

    action = f"magnus execute_single_node {run_id} " f"{node._command_friendly_name()}" f" --log-level {log_level}"

    if context.run_context.pipeline_file:
        action = action + f" --file {context.run_context.pipeline_file}"

    if map_variable:
        action = action + f" --map-variable '{json.dumps(map_variable)}'"

    if context.run_context.configuration_file:
        action = action + f" --config-file {context.run_context.configuration_file}"

    if context.run_context.parameters_file:
        action = action + f" --parameters-file {context.run_context.parameters_file}"

    if context.run_context.tag:
        action = action + f" --tag {context.run_context.tag}"

    return action


def get_fan_command(
    mode: str,
    node: BaseNode,
    run_id: str,
    map_variable: TypeMapVariable = None,
) -> str:
    """
    An utility function to return the fan "in or out" command

    Args:
        executor (BaseExecutor): The executor class
        mode (str): in or out
        node (BaseNode): The composite node that we are fanning in or out
        run_id (str): The run id.
        map_variable (dict, optional): If the node is a map, we have the map variable. Defaults to None.

    Returns:
        str: The fan in or out command
    """
    log_level = logging.getLevelName(logger.getEffectiveLevel())
    action = (
        f"magnus fan {run_id} "
        f"{node._command_friendly_name()} "
        f"--mode {mode} "
        f"--file {context.run_context.pipeline_file} "
        f"--log-level {log_level} "
    )
    if context.run_context.configuration_file:
        action = action + f" --config-file {context.run_context.configuration_file} "

    if context.run_context.parameters_file:
        action = action + f" --parameters-file {context.run_context.parameters_file}"

    if map_variable:
        action = action + f" --map-variable '{json.dumps(map_variable)}'"

    if context.run_context.tag:
        action = action + f" --tag {context.run_context.tag}"

    return action


def get_job_execution_command(node: TaskNode, over_write_run_id: str = "") -> str:
    """Get the execution command to run a job via command line.

    This function should be used by all executors to submit jobs in remote environment

    Args:
        executor (BaseExecutor): The executor class.
        node (BaseNode): The node being executed.
        over_write_run_id (str, optional): If the node is part of a map step. Defaults to ''.

    Returns:
        str: The execution command to run a job via command line.
    """

    run_id = context.run_context.run_id

    if over_write_run_id:
        run_id = over_write_run_id

    log_level = logging.getLevelName(logger.getEffectiveLevel())

    cli_command, cli_options = node.executable.get_cli_options()

    action = f"magnus execute_{cli_command} {run_id} " f" --log-level {log_level}"

    action = action + f" --entrypoint {defaults.ENTRYPOINT.SYSTEM.value}"

    if context.run_context.configuration_file:
        action = action + f" --config-file {context.run_context.configuration_file}"

    if context.run_context.parameters_file:
        action = action + f" --parameters-file {context.run_context.parameters_file}"

    if context.run_context.tag:
        action = action + f" --tag {context.run_context.tag}"

    for key, value in cli_options.items():
        action = action + f" --{key} {value}"

    return action


def get_provider_by_name_and_type(service_type: str, service_details: defaults.ServiceConfig):
    """Given a service type, one of executor, run_log_store, catalog, secrets and the config
    return the exact child class implementing the service.
    We use stevedore to do the work for us.

    Args:
        service_type (str): One of executor, run_log_store, catalog, secrets
        service_details (dict): The config used to instantiate the service.

    Raises:
        Exception: If the service by that name does not exist

    Returns:
        object: A service object
    """
    namespace = service_type

    service_name = service_details["type"]
    service_config: Mapping[str, Any] = {}
    if "config" in service_details:
        service_config = service_details.get("config", {})

    logger.info(f"Trying to get a service of {service_type} of the name {service_name} with config: {service_config}")
    try:
        mgr = driver.DriverManager(
            namespace=namespace,
            name=service_name,
            invoke_on_load=True,
            invoke_kwds={**service_config},
        )
        return mgr.driver
    except Exception as _e:
        raise Exception(f"Could not find the service of type: {service_type} with config: {service_details}") from _e


def get_duration_between_datetime_strings(start_time: str, end_time: str) -> str:
    """Given two datetime strings, compute the duration between them.

    Args:
        start_time (str): ISO format datetime string
        end_time (str): ISO format datetime string
    Returns:
        The duration between the time in string format
    """
    start = datetime.fromisoformat(start_time)
    end = datetime.fromisoformat(end_time)

    return str(end - start)


def get_run_config() -> dict:
    """Given an executor with assigned services, return the run_config.

    Args:
        executor (object): The executor with all the services assigned.

    Returns:
        dict: The run_config.
    """

    run_config = context.run_context.model_dump(by_alias=True)
    return run_config


def json_to_ordered_dict(json_str: str) -> TypeMapVariable:
    """Decode a JSON str into OrderedDict.

    Args:
        json_str ([str]): The JSON string to decode

    Returns:
        [OrderedDict]: The decoded OrderedDict
    """
    if json_str and json_str != "{}":
        return json.loads(json_str, object_pairs_hook=OrderedDict)

    return OrderedDict()


def set_magnus_environment_variables(run_id: str = "", configuration_file: str = "", tag: str = "") -> None:
    """Set the environment variables used by magnus. This function should be called during the prepare configurations
    by all executors.

    Args:
        run_id (str, optional): The run id of the execution. Defaults to None.
        configuration_file (str, optional): The configuration file if used. Defaults to None.
        tag (str, optional): The tag associated with a run. Defaults to None.
    """
    if run_id:
        os.environ[defaults.ENV_RUN_ID] = run_id

    if configuration_file:
        os.environ[defaults.MAGNUS_CONFIG_FILE] = configuration_file

    if tag:
        os.environ[defaults.MAGNUS_RUN_TAG] = tag


def gather_variables() -> dict:
    """Gather all the environment variables used by magnus. All the variables start with MAGNUS_VAR_.

    Returns:
        dict: All the environment variables present in the environment.
    """
    variables = {}

    for env_var, value in os.environ.items():
        if env_var.startswith(defaults.VARIABLE_PREFIX):
            key = remove_prefix(env_var, defaults.VARIABLE_PREFIX)
            variables[key] = value

    return variables



import pickle
from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel, ConfigDict

import magnus.context as context


class BasePickler(ABC, BaseModel):
    """
    The base class for all picklers.

    We are still in the process of hardening the design of this class.
    For now, we are just going to use pickle.
    """

    extension: str = ""
    service_name: str = ""
    service_type: str = "pickler"
    model_config = ConfigDict(extra="forbid")

    @property
    def _context(self):
        return context.run_context

    @abstractmethod
    def dump(self, data: Any, path: str):
        """
        Dump an object to the specified path.
        The path is the full path.

        To correctly identify the pickler from possible implementations, we use the extension.
        An extension is added automatically, if not provided.

        Args:
            data (Any): The object to pickle
            path (str): The path to save the pickle file

        Raises:
            NotImplementedError: Base class has no implementation
        """
        raise NotImplementedError

    @abstractmethod
    def load(self, path: str) -> Any:
        """
        Load the object from the specified path.

        To correctly identify the pickler from possible implementations, we use the extension.
        An extension is added automatically, if not provided.

        Args:
            path (str): The path to load the pickled file from.

        Raises:
            NotImplementedError: Base class has no implementation.
        """
        raise NotImplementedError


class NativePickler(BasePickler):
    """
    Uses native python pickle to load and dump files
    """

    extension: str = ".pickle"
    service_name: str = "pickle"

    def dump(self, data: Any, path: str):
        """
        Dump an object to the specified path.
        The path is the full path.

        Args:
            data (Any): The data to pickle
            path (str): The path to save the pickle file
        """
        if not path.endswith(self.extension):
            path = path + self.extension

        with open(path, "wb") as f:
            pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)

    def load(self, path: str) -> Any:
        """
        Load the object from the specified path.

        Args:
            path (str): The path to load the object from.

        Returns:
            Any: The data loaded from the file.
        """
        if not path.endswith(self.extension):
            path = path + self.extension

        data = None
        with open(path, "rb") as f:
            data = pickle.load(f)

        return data



# import logging
# import shlex
# from typing import Dict, List, Optional

# from pydantic import BaseModel

# from magnus import defaults, integration, utils
# from magnus.executor import BaseExecutor
# from magnus.graph import Graph
# from magnus.nodes import BaseNode

# logger = logging.getLogger(defaults.NAME)

# try:
#     from kubernetes import client
#     from kubernetes.client import V1EnvVar, V1EnvVarSource, V1PersistentVolumeClaimVolumeSource, V1SecretKeySelector
# except ImportError as _e:
#     msg = "Kubernetes Dependencies have not been installed!!"
#     # raise Exception(msg) from _e


# class Toleration(BaseModel):
#     effect: str
#     key: str
#     operator: str
#     value: str


# class K8sJobExecutor(BaseExecutor):
#     service_name = "k8s-job"

#     # TODO: move this to K8's style config.
#     class ContextConfig(BaseModel):
#         docker_image: str
#         config_path: str = ""  # Let the client decide on the path to the config file.
#         namespace: str = "default"
#         cpu_limit: str = "250m"
#         memory_limit: str = "1G"
#         gpu_limit: int = 0
#         gpu_vendor: str = "nvidia.com/gpu"
#         cpu_request: str = ""
#         memory_request: str = ""
#         active_deadline_seconds: int = 60 * 60 * 2  # 2 hours
#         ttl_seconds_after_finished: int = 60  #  1 minute
#         image_pull_policy: str = "Always"
#         secrets_from_k8s: dict = {}  # EnvVar=SecretName:Key
#         persistent_volumes: dict = {}  # volume-name:mount_path
#         node_selector: Dict[str, str] = {}
#         tolerations: List[Toleration] = []
#         labels: Dict[str, str] = {}

#     def __init__(self, config: Optional[dict] = None):
#         self.config = self.ContextConfig(**(config or {}))
#         self.persistent_volumes = {}

#         for i, (claim, mount_path) in enumerate(self.config.persistent_volumes.items()):
#             self.persistent_volumes[f"executor-{i}"] = (claim, mount_path)

#     def prepare_for_graph_execution(self):
#         """
#         This method would be called prior to calling execute_graph.
#         Perform any steps required before doing the graph execution.

#         The most common implementation is to prepare a run log for the run if the run uses local interactive compute.

#         But in cases of actual rendering the job specs (eg: AWS step functions, K8's) we need not do anything.
#         """

#         integration.validate(self, self.run_log_store)
#         integration.configure_for_traversal(self, self.run_log_store)

#         integration.validate(self, self.catalog_handler)
#         integration.configure_for_traversal(self, self.catalog_handler)

#         integration.validate(self, self.secrets_handler)
#         integration.configure_for_traversal(self, self.secrets_handler)

#         integration.validate(self, self.experiment_tracker)
#         integration.configure_for_traversal(self, self.experiment_tracker)

#     def prepare_for_node_execution(self):
#         """
#         Perform any modifications to the services prior to execution of the node.

#         Args:
#             node (Node): [description]
#             map_variable (dict, optional): [description]. Defaults to None.
#         """

#         integration.validate(self, self.run_log_store)
#         integration.configure_for_execution(self, self.run_log_store)

#         integration.validate(self, self.catalog_handler)
#         integration.configure_for_execution(self, self.catalog_handler)

#         integration.validate(self, self.secrets_handler)
#         integration.configure_for_execution(self, self.secrets_handler)

#         integration.validate(self, self.experiment_tracker)
#         integration.configure_for_execution(self, self.experiment_tracker)

#         self._set_up_run_log(exists_ok=True)

#     @property
#     def _client(self):
#         from kubernetes import config as k8s_config

#         if self.config.config_path:
#             k8s_config.load_kube_config(kube_config_path=self.config.config_path)
#         else:
#             # https://github.com/kubernetes-client/python/blob/master/kubernetes/base/config/__init__.py
#             k8s_config.load_config()
#         return client

#     @property
#     def tolerations(self):
#         return [toleration.dict() for toleration in self.config.tolerations]

#     def execute_job(self, node: BaseNode):
#         command = utils.get_job_execution_command(self, node)
#         logger.info(f"Triggering a kubernetes job with : {command}")

#         self.config.labels["job_name"] = self.run_id

#         k8s_batch = self._client.BatchV1Api()

#         cpu_limit = self.config.cpu_limit
#         memory_limit = self.config.memory_limit

#         cpu_request = self.config.cpu_request or cpu_limit
#         memory_request = self.config.memory_request or memory_limit

#         gpu_limit = str(self.config.gpu_limit)  # Should be something like nvidia -etc

#         limits = {
#             "cpu": cpu_limit,
#             "memory": memory_limit,
#             self.config.gpu_vendor: gpu_limit,
#         }
#         requests = {"cpu": cpu_request, "memory": memory_request}
#         resources = {"limits": limits, "requests": requests}

#         environment_variables = []
#         for secret_env, k8_secret in self.config.secrets_from_k8s.items():
#             try:
#                 secret_name, key = k8_secret.split(":")
#             except Exception as _e:
#                 msg = "K8's secret should be of format EnvVar=SecretName:Key"
#                 raise Exception(msg) from _e
#             secret_as_env = V1EnvVar(
#                 name=secret_env,
#                 value_from=V1EnvVarSource(secret_key_ref=V1SecretKeySelector(name=secret_name, key=key)),
#             )
#             environment_variables.append(secret_as_env)

#         overridden_params = utils.get_user_set_parameters()
#         # The parameters present in the environment override the parameters present in the parameters file
#         # The values are coerced to be strings, hopefully they will be fine on the other side.
#         for k, v in overridden_params.items():
#             environment_variables.append(V1EnvVar(name=defaults.PARAMETER_PREFIX + k, value=str(v)))

#         pod_volumes = []
#         volume_mounts = []
#         for claim_name, (claim, mount_path) in self.persistent_volumes.items():
#             pod_volumes.append(
#                 self._client.V1Volume(
#                     name=claim_name,
#                     persistent_volume_claim=V1PersistentVolumeClaimVolumeSource(claim_name=claim),
#                 )
#             )
#             volume_mounts.append(self._client.V1VolumeMount(name=claim_name, mount_path=mount_path))

#         base_container = self._client.V1Container(
#             name=self.run_id,
#             image=self.config.docker_image,
#             command=shlex.split(command),
#             resources=resources,
#             env=environment_variables,
#             image_pull_policy="Always",
#             volume_mounts=volume_mounts or None,
#         )

#         pod_spec = self._client.V1PodSpec(
#             volumes=pod_volumes or None,
#             restart_policy="Never",
#             containers=[base_container],
#             node_selector=self.config.node_selector,
#             tolerations=self.tolerations,
#         )

#         pod_template = self._client.V1PodTemplateSpec(
#             metadata=client.V1ObjectMeta(
#                 labels=self.config.labels,
#                 annotations={"sidecar.istio.io/inject": "false"},
#             ),
#             spec=pod_spec,
#         )

#         job_spec = client.V1JobSpec(
#             template=pod_template,
#             backoff_limit=2,
#             ttl_seconds_after_finished=self.config.ttl_seconds_after_finished,
#         )
#         job_spec.active_deadline_seconds = self.config.active_deadline_seconds

#         job = client.V1Job(
#             api_version="batch/v1",
#             kind="Job",
#             metadata=client.V1ObjectMeta(name=self.run_id),
#             spec=job_spec,
#         )

#         logger.debug(f"Submitting kubernetes job: {job}")

#         try:
#             response = k8s_batch.create_namespaced_job(
#                 body=job,
#                 namespace=self.config.namespace,
#                 _preload_content=False,
#                 pretty=True,
#             )
#             print(f"Kubernetes job {self.run_id} created")
#             logger.debug(f"Kubernetes job response: {response}")
#         except Exception as e:
#             logger.exception(e)
#             raise

#     def execute_node(self, node: BaseNode, map_variable: Optional[dict] = None, **kwargs):
#         step_log = self.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable))

#         self.add_code_identities(node=node, step_log=step_log)

#         step_log.step_type = node.node_type
#         step_log.status = defaults.PROCESSING
#         self.run_log_store.add_step_log(step_log, self.run_id)

#         super()._execute_node(node, map_variable=map_variable, **kwargs)

#         step_log = self.run_log_store.get_step_log(node._get_step_log_name(map_variable), self.run_id)
#         if step_log.status == defaults.FAIL:
#             raise Exception(f"Step {node.name} failed")

#     def execute_graph(self, dag: Graph, map_variable: Optional[dict] = None, **kwargs):
#         msg = "This executor is not supported to execute any graphs but only jobs (functions or notebooks)"
#         raise NotImplementedError(msg)

#     def send_return_code(self, stage="traversal"):
#         """
#         Convenience function used by pipeline to send return code to the caller of the cli

#         Raises:
#             Exception: If the pipeline execution failed
#         """
#         if stage != "traversal":  # traversal does no actual execution, so return code is pointless
#             run_id = self.run_id

#             run_log = self.run_log_store.get_run_log_by_id(run_id=run_id, full=False)
#             if run_log.status == defaults.FAIL:
#                 raise Exception("Pipeline execution failed")



import json
import logging
from typing import Optional, cast

from rich import print

import magnus.context as context
from magnus import defaults, graph, utils
from magnus.defaults import MagnusConfig, ServiceConfig

logger = logging.getLogger(defaults.LOGGER_NAME)


def get_default_configs() -> MagnusConfig:
    """
    User can provide extensions as part of their code base, magnus-config.yaml provides the place to put them.
    """
    user_configs = {}
    if utils.does_file_exist(defaults.USER_CONFIG_FILE):
        user_configs = utils.load_yaml(defaults.USER_CONFIG_FILE)

    if not user_configs:
        return {}

    user_defaults = user_configs.get("defaults", {})
    if user_defaults:
        return user_defaults

    return {}


def prepare_configurations(
    run_id: str,
    configuration_file: str = "",
    pipeline_file: str = "",
    tag: str = "",
    use_cached: str = "",
    parameters_file: str = "",
    force_local_executor: bool = False,
) -> context.Context:
    """
    Replace the placeholders in the dag/config against the variables file.

    Attach the secrets_handler, run_log_store, catalog_handler to the executor and return it.

    Args:
        variables_file (str): The variables file, if used or None
        pipeline_file (str): The config/dag file
        run_id (str): The run id of the run.
        tag (str): If a tag is provided at the run time
        use_cached (str): Provide the run_id of the older run

    Returns:
        executor.BaseExecutor : A prepared executor as per the dag/config
    """
    magnus_defaults = get_default_configs()

    variables = utils.gather_variables()

    templated_configuration = {}
    if configuration_file:
        templated_configuration = utils.load_yaml(configuration_file) or {}

    configuration: MagnusConfig = cast(MagnusConfig, templated_configuration)

    # Run log settings, configuration over-rides everything
    run_log_config: Optional[ServiceConfig] = configuration.get("run_log_store", None)
    if not run_log_config:
        run_log_config = cast(ServiceConfig, magnus_defaults.get("run_log_store", defaults.DEFAULT_RUN_LOG_STORE))
    run_log_store = utils.get_provider_by_name_and_type("run_log_store", run_log_config)

    # Catalog handler settings, configuration over-rides everything
    catalog_config: Optional[ServiceConfig] = configuration.get("catalog", None)
    if not catalog_config:
        catalog_config = cast(ServiceConfig, magnus_defaults.get("catalog", defaults.DEFAULT_CATALOG))
    catalog_handler = utils.get_provider_by_name_and_type("catalog", catalog_config)

    # Secret handler settings, configuration over-rides everything
    secrets_config: Optional[ServiceConfig] = configuration.get("secrets", None)
    if not secrets_config:
        secrets_config = cast(ServiceConfig, magnus_defaults.get("secrets", defaults.DEFAULT_SECRETS))
    secrets_handler = utils.get_provider_by_name_and_type("secrets", secrets_config)

    # experiment tracker settings, configuration over-rides everything
    tracker_config: Optional[ServiceConfig] = configuration.get("experiment_tracker", None)
    if not tracker_config:
        tracker_config = cast(
            ServiceConfig, magnus_defaults.get("experiment_tracker", defaults.DEFAULT_EXPERIMENT_TRACKER)
        )
    tracker_handler = utils.get_provider_by_name_and_type("experiment_tracker", tracker_config)

    # executor configurations, configuration over rides everything
    executor_config: Optional[ServiceConfig] = configuration.get("executor", None)
    if force_local_executor:
        executor_config = ServiceConfig(type="local", config={})

    if not executor_config:
        executor_config = cast(ServiceConfig, magnus_defaults.get("executor", defaults.DEFAULT_EXECUTOR))
    configured_executor = utils.get_provider_by_name_and_type("executor", executor_config)

    # Construct the context
    run_context = context.Context(
        executor=configured_executor,
        run_log_store=run_log_store,
        catalog_handler=catalog_handler,
        secrets_handler=secrets_handler,
        experiment_tracker=tracker_handler,
        variables=variables,
        tag=tag,
        run_id=run_id,
        configuration_file=configuration_file,
        parameters_file=parameters_file,
    )

    if pipeline_file:
        # There are use cases where we are only preparing the executor
        pipeline_config = utils.load_yaml(pipeline_file)

        logger.info("The input pipeline:")
        logger.info(json.dumps(pipeline_config, indent=4))

        # Create the graph
        dag_config = pipeline_config["dag"]
        dag_hash = utils.get_dag_hash(dag_config)
        dag = graph.create_graph(dag_config)

        run_context.pipeline_file = pipeline_file
        run_context.dag = dag
        run_context.dag_hash = dag_hash

    run_context.use_cached = False
    if use_cached:
        run_context.use_cached = True
        run_context.original_run_id = use_cached

    context.run_context = run_context

    return run_context


def execute(
    configuration_file: str,
    pipeline_file: str,
    tag: str = "",
    run_id: str = "",
    use_cached: str = "",
    parameters_file: str = "",
):
    # pylint: disable=R0914,R0913
    """
    The entry point to magnus execution. This method would prepare the configurations and delegates traversal to the
    executor

    Args:
        pipeline_file (str): The config/dag file
        run_id (str): The run id of the run.
        tag (str): If a tag is provided at the run time
        use_cached (str): The previous run_id to use.
        parameters_file (str): The parameters being sent in to the application
    """
    # Re run settings
    run_id = utils.generate_run_id(run_id=run_id)

    run_context = prepare_configurations(
        configuration_file=configuration_file,
        pipeline_file=pipeline_file,
        run_id=run_id,
        tag=tag,
        use_cached=use_cached,
        parameters_file=parameters_file,
    )
    print("Working with context:")
    print(run_context)

    executor = run_context.executor

    run_context.execution_plan = defaults.EXECUTION_PLAN.CHAINED.value

    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    # Prepare for graph execution
    executor.prepare_for_graph_execution()

    logger.info("Executing the graph")
    executor.execute_graph(dag=run_context.dag)  # type: ignore

    executor.send_return_code()


def execute_single_node(
    configuration_file: str,
    pipeline_file: str,
    step_name: str,
    map_variable: str,
    run_id: str,
    tag: str = "",
    parameters_file: str = "",
):
    """
    The entry point into executing a single node of magnus. Orchestration modes should extensively use this
    entry point.

    It should have similar set up of configurations to execute because orchestrator modes can initiate the execution.

    Args:
        variables_file (str): The variables file, if used or None
        step_name : The name of the step to execute in dot path convention
        pipeline_file (str): The config/dag file
        run_id (str): The run id of the run.
        tag (str): If a tag is provided at the run time
        parameters_file (str): The parameters being sent in to the application

    """
    from magnus import nodes

    run_context = prepare_configurations(
        configuration_file=configuration_file,
        pipeline_file=pipeline_file,
        run_id=run_id,
        tag=tag,
        use_cached="",
        parameters_file=parameters_file,
    )
    print("Working with context:")
    print(run_context)

    executor = run_context.executor
    run_context.execution_plan = defaults.EXECUTION_PLAN.CHAINED.value
    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    executor.prepare_for_node_execution()

    if not run_context.dag:
        # There are a few entry points that make graph dynamically and do not have a dag defined statically.
        run_log = run_context.run_log_store.get_run_log_by_id(run_id=run_id, full=False)
        run_context.dag = graph.create_graph(run_log.run_config["pipeline"])

    step_internal_name = nodes.BaseNode._get_internal_name_from_command_name(step_name)

    map_variable_dict = utils.json_to_ordered_dict(map_variable)

    node_to_execute, _ = graph.search_node_by_internal_name(run_context.dag, step_internal_name)

    logger.info("Executing the single node of : %s", node_to_execute)
    executor.execute_node(node=node_to_execute, map_variable=map_variable_dict)

    executor.send_return_code(stage="execution")


def execute_single_brach(
    configuration_file: str,
    pipeline_file: str,
    branch_name: str,
    map_variable: str,
    run_id: str,
    tag: str,
):
    """
    The entry point into executing a branch of the graph. Interactive modes in parallel runs use this to execute
    branches in parallel.

    This entry point is never used by its own but rather from a node. So the arguments sent into this are fewer.

    Args:
        variables_file (str): The variables file, if used or None
        branch_name : The name of the branch to execute, in dot.path.convention
        pipeline_file (str): The config/dag file
        run_id (str): The run id of the run.
        tag (str): If a tag is provided at the run time
    """
    from magnus import nodes

    run_context = prepare_configurations(
        configuration_file=configuration_file,
        pipeline_file=pipeline_file,
        run_id=run_id,
        tag=tag,
        use_cached="",
    )
    print("Working with context:")
    print(run_context)

    executor = run_context.executor
    run_context.execution_plan = defaults.EXECUTION_PLAN.CHAINED.value
    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    branch_internal_name = nodes.BaseNode._get_internal_name_from_command_name(branch_name)

    map_variable_dict = utils.json_to_ordered_dict(map_variable)

    branch_to_execute = graph.search_branch_by_internal_name(run_context.dag, branch_internal_name)  # type: ignore

    logger.info("Executing the single branch of %s", branch_to_execute)
    executor.execute_graph(dag=branch_to_execute, map_variable=map_variable_dict)

    executor.send_return_code()


def execute_notebook(
    entrypoint: str,
    notebook_file: str,
    catalog_config: dict,
    configuration_file: str,
    notebook_output_path: str = "",
    tag: str = "",
    run_id: str = "",
    parameters_file: str = "",
):
    """
    The entry point to magnus execution of a notebook. This method would prepare the configurations and
    delegates traversal to the executor
    """
    run_id = utils.generate_run_id(run_id=run_id)

    run_context = prepare_configurations(
        configuration_file=configuration_file,
        run_id=run_id,
        tag=tag,
        parameters_file=parameters_file,
    )

    executor = run_context.executor
    run_context.execution_plan = defaults.EXECUTION_PLAN.UNCHAINED.value
    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    print("Working with context:")
    print(run_context)

    step_config = {
        "command": notebook_file,
        "command_type": "notebook",
        "notebook_output_path": notebook_output_path,
        "type": "task",
        "next": "success",
        "catalog": catalog_config,
    }
    node = graph.create_node(name="executing job", step_config=step_config)

    if entrypoint == defaults.ENTRYPOINT.USER.value:
        # Prepare for graph execution
        executor.prepare_for_graph_execution()

        logger.info("Executing the job from the user. We are still in the caller's compute environment")
        executor.execute_job(node=node)

    elif entrypoint == defaults.ENTRYPOINT.SYSTEM.value:
        executor.prepare_for_node_execution()
        logger.info("Executing the job from the system. We are in the config's compute environment")
        executor.execute_node(node=node)

        # Update the status of the run log
        step_log = run_context.run_log_store.get_step_log(node._get_step_log_name(), run_id)
        run_context.run_log_store.update_run_log_status(run_id=run_id, status=step_log.status)

    else:
        raise ValueError(f"Invalid entrypoint {entrypoint}")

    executor.send_return_code()


def execute_function(
    entrypoint: str,
    command: str,
    catalog_config: dict,
    configuration_file: str,
    tag: str = "",
    run_id: str = "",
    parameters_file: str = "",
):
    """
    The entry point to magnus execution of a function. This method would prepare the configurations and
    delegates traversal to the executor
    """
    run_id = utils.generate_run_id(run_id=run_id)

    run_context = prepare_configurations(
        configuration_file=configuration_file,
        run_id=run_id,
        tag=tag,
        parameters_file=parameters_file,
    )

    executor = run_context.executor

    run_context.execution_plan = defaults.EXECUTION_PLAN.UNCHAINED.value
    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    print("Working with context:")
    print(run_context)

    # Prepare the graph with a single node
    step_config = {
        "command": command,
        "command_type": "python",
        "type": "task",
        "next": "success",
        "catalog": catalog_config,
    }
    node = graph.create_node(name="executing job", step_config=step_config)

    if entrypoint == defaults.ENTRYPOINT.USER.value:
        # Prepare for graph execution
        executor.prepare_for_graph_execution()

        logger.info("Executing the job from the user. We are still in the caller's compute environment")
        executor.execute_job(node=node)

    elif entrypoint == defaults.ENTRYPOINT.SYSTEM.value:
        executor.prepare_for_node_execution()
        logger.info("Executing the job from the system. We are in the config's compute environment")
        executor.execute_node(node=node)

        # Update the status of the run log
        step_log = run_context.run_log_store.get_step_log(node._get_step_log_name(), run_id)
        run_context.run_log_store.update_run_log_status(run_id=run_id, status=step_log.status)

    else:
        raise ValueError(f"Invalid entrypoint {entrypoint}")

    executor.send_return_code()


def fan(
    configuration_file: str,
    pipeline_file: str,
    step_name: str,
    mode: str,
    map_variable: str,
    run_id: str,
    tag: str = "",
    parameters_file: str = "",
):
    """
    The entry point to either fan in or out for a composite node. Only 3rd party orchestrators should use this.

    It should have similar set up of configurations to execute because orchestrator modes can initiate the execution.

    Args:
        configuration_file (str): The configuration file.
        mode: in or out
        step_name : The name of the step to execute in dot path convention
        pipeline_file (str): The config/dag file
        run_id (str): The run id of the run.
        tag (str): If a tag is provided at the run time
        parameters_file (str): The parameters being sent in to the application

    """
    from magnus import nodes

    run_context = prepare_configurations(
        configuration_file=configuration_file,
        pipeline_file=pipeline_file,
        run_id=run_id,
        tag=tag,
        use_cached="",
        parameters_file=parameters_file,
    )
    print("Working with context:")
    print(run_context)

    executor = run_context.executor
    run_context.execution_plan = defaults.EXECUTION_PLAN.CHAINED.value
    utils.set_magnus_environment_variables(run_id=run_id, configuration_file=configuration_file, tag=tag)

    executor.prepare_for_node_execution()

    step_internal_name = nodes.BaseNode._get_internal_name_from_command_name(step_name)
    node_to_execute, _ = graph.search_node_by_internal_name(run_context.dag, step_internal_name)  # type: ignore

    map_variable_dict = utils.json_to_ordered_dict(map_variable)

    if mode == "in":
        logger.info("Fanning in for : %s", node_to_execute)
        executor.fan_in(node=node_to_execute, map_variable=map_variable_dict)
    elif mode == "out":
        logger.info("Fanning out for : %s", node_to_execute)
        executor.fan_out(node=node_to_execute, map_variable=map_variable_dict)
    else:
        raise ValueError(f"Invalid mode {mode}")


if __name__ == "__main__":
    # This is only for perf testing purposes.
    prepare_configurations(run_id="abc", pipeline_file="example/mocking.yaml")



import logging
from pathlib import Path

from magnus import defaults
from magnus.extensions.catalog.file_system.implementation import FileSystemCatalog

logger = logging.getLogger(defaults.LOGGER_NAME)


class K8sPVCatalog(FileSystemCatalog):
    service_name: str = "k8s-pvc"
    persistent_volume_name: str
    mount_path: str

    def get_catalog_location(self):
        return str(Path(self.mount_path) / self.catalog_location)



from __future__ import annotations

import logging
import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict

import magnus.context as context
from magnus import defaults
from magnus.datastore import DataCatalog, RunLog, StepLog
from magnus.defaults import TypeMapVariable
from magnus.graph import Graph

if TYPE_CHECKING:  # pragma: no cover
    from magnus.extensions.nodes import TaskNode
    from magnus.nodes import BaseNode

logger = logging.getLogger(defaults.LOGGER_NAME)


class BaseExecutor(ABC, BaseModel):
    """
    The skeleton of an executor class.
    Any implementation of an executor should inherit this class and over-ride accordingly.

    There is a extension available in magnus/extensions/executor/__init__.py
    which implements the most common functionality which is easier to
    extend/override in most scenarios.

    """

    service_name: str = ""
    service_type: str = "executor"

    enable_parallel: bool = defaults.ENABLE_PARALLEL
    overrides: dict = {}

    _previous_run_log: Optional[RunLog] = None
    _single_step: str = ""

    _context_step_log = None  # type : StepLog
    _context_node = None  # type: BaseNode
    model_config = ConfigDict(extra="forbid")

    @property
    def _context(self):
        return context.run_context

    def _is_parallel_execution(self) -> bool:
        """
        Controls the parallelization of branches in map and parallel state.
        Defaults to False and left for the compute modes to decide.

        Interactive executors like local and local-container need decisions.
        For most transpilers it is inconsequential as its always True and supported by platforms.

        Returns:
            bool: True if the execution allows parallel execution of branches.
        """
        return self.enable_parallel

    @abstractmethod
    def _get_parameters(self) -> Dict[str, Any]:
        """
        Get the parameters for the execution.
        The parameters can be defined in parameters file and can be overridden by environment variables.

        Returns:
            Dict[str, Any]: The parameters for the execution.
        """
        ...

    @abstractmethod
    def _set_up_for_re_run(self, parameters: Dict[str, Any]) -> None:
        """
        Set up the executor for using a previous execution.

        Retrieve the older run log, error out if it does not exist.
        Sync the catalogs from the previous run log with the current one.

        Update the parameters of this execution with the previous one. The previous one take precedence.

        Args:
            parameters (Dict[str, Any]): The parameters for the current execution.
        """

    @abstractmethod
    def _set_up_run_log(self, exists_ok=False):
        """
        Create a run log and put that in the run log store

        If exists_ok, we allow the run log to be already present in the run log store.
        """
        ...

    @abstractmethod
    def prepare_for_graph_execution(self):
        """
        This method should be called prior to calling execute_graph.
        Perform any steps required before doing the graph execution.

        The most common implementation is to prepare a run log for the run if the run uses local interactive compute.

        But in cases of actual rendering the job specs (eg: AWS step functions, K8's) we check if the services are OK.
        We do not set up a run log as its not relevant.
        """
        ...

    @abstractmethod
    def prepare_for_node_execution(self):
        """
        Perform any modifications to the services prior to execution of the node.

        Args:
            node (Node): [description]
            map_variable (dict, optional): [description]. Defaults to None.
        """
        ...

    @abstractmethod
    def _sync_catalog(self, step_log: StepLog, stage: str, synced_catalogs=None) -> Optional[List[DataCatalog]]:
        """
        1). Identify the catalog settings by over-riding node settings with the global settings.
        2). For stage = get:
                Identify the catalog items that are being asked to get from the catalog
                And copy them to the local compute data folder
        3). For stage = put:
                Identify the catalog items that are being asked to put into the catalog
                Copy the items from local compute folder to the catalog
        4). Add the items onto the step log according to the stage

        Args:
            node (Node): The current node being processed
            step_log (StepLog): The step log corresponding to that node
            stage (str): One of get or put

        Raises:
            Exception: If the stage is not in one of get/put

        """
        ...

    @abstractmethod
    def get_effective_compute_data_folder(self) -> Optional[str]:
        """
        Get the effective compute data folder for the given stage.
        If there is nothing to catalog, we return None.

        The default is the compute data folder of the catalog but this can be over-ridden by the node.

        Args:
            stage (str): The stage we are in the process of cataloging


        Returns:
            Optional[str]: The compute data folder as defined by catalog handler or the node or None.
        """
        ...

    @property
    def step_attempt_number(self) -> int:
        """
        The attempt number of the current step.
        Orchestrators should use this step to submit multiple attempts of the job.

        Returns:
            int: The attempt number of the current step. Defaults to 1.
        """
        return int(os.environ.get(defaults.ATTEMPT_NUMBER, 1))

    @abstractmethod
    def _execute_node(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        This is the entry point when we do the actual execution of the function.

        While in interactive execution, we just compute, in 3rd party interactive execution, we need to reach
        this function.

        In most cases,
            * We get the corresponding step_log of the node and the parameters.
            * We sync the catalog to GET any data sets that are in the catalog
            * We call the execute method of the node for the actual compute and retry it as many times as asked.
            * If the node succeeds, we get any of the user defined metrics provided by the user.
            * We sync the catalog to PUT any data sets that are in the catalog.

        Args:
            node (Node): The node to execute
            map_variable (dict, optional): If the node is of a map state, map_variable is the value of the iterable.
                        Defaults to None.
        """
        ...

    @abstractmethod
    def execute_node(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        The entry point for all executors apart from local.
        We have already prepared for node execution.

        Args:
            node (BaseNode): The node to execute
            map_variable (dict, optional): If the node is part of a map, send in the map dictionary. Defaults to None.

        Raises:
            NotImplementedError: _description_
        """
        ...

    @abstractmethod
    def add_code_identities(self, node: BaseNode, step_log: StepLog, **kwargs):
        """
        Add code identities specific to the implementation.

        The Base class has an implementation of adding git code identities.

        Args:
            step_log (object): The step log object
            node (BaseNode): The node we are adding the step log for
        """
        ...

    @abstractmethod
    def execute_from_graph(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        This is the entry point to from the graph execution.

        While the self.execute_graph is responsible for traversing the graph, this function is responsible for
        actual execution of the node.

        If the node type is:
            * task : We can delegate to _execute_node after checking the eligibility for re-run in cases of a re-run
            * success: We can delegate to _execute_node
            * fail: We can delegate to _execute_node

        For nodes that are internally graphs:
            * parallel: Delegate the responsibility of execution to the node.execute_as_graph()
            * dag: Delegate the responsibility of execution to the node.execute_as_graph()
            * map: Delegate the responsibility of execution to the node.execute_as_graph()

        Transpilers will NEVER use this method and will NEVER call ths method.
        This method should only be used by interactive executors.

        Args:
            node (Node): The node to execute
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of iterable.
                    Defaults to None.
        """
        ...

    @abstractmethod
    def trigger_job(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        Executor specific way of triggering jobs when magnus does both traversal and execution

        Transpilers will NEVER use this method and will NEVER call them.
        Only interactive executors who need execute_from_graph will ever implement it.

        Args:
            node (BaseNode): The node to execute
            map_variable (str, optional): If the node if of a map state, this corresponds to the value of iterable.
                    Defaults to ''.

        NOTE: We do not raise an exception as this method is not required by many extensions
        """
        ...

    @abstractmethod
    def _get_status_and_next_node_name(self, current_node: BaseNode, dag: Graph, map_variable: TypeMapVariable = None):
        """
        Given the current node and the graph, returns the name of the next node to execute.

        The name is always relative the graph that the node resides in.

        If the current node succeeded, we return the next node as per the graph.
        If the current node failed, we return the on failure node of the node (if provided) or the global one.

        Args:
            current_node (BaseNode): The current node.
            dag (Graph): The dag we are traversing.
            map_variable (dict): If the node belongs to a map branch.
        """

        ...

    @abstractmethod
    def execute_graph(self, dag: Graph, map_variable: TypeMapVariable = None, **kwargs):
        """
        The parallelization is controlled by the nodes and not by this function.

        Transpilers should over ride this method to do the translation of dag to the platform specific way.
        Interactive methods should use this to traverse and execute the dag.
            - Use execute_from_graph to handle sub-graphs

        Logically the method should:
            * Start at the dag.start_at of the dag.
            * Call the self.execute_from_graph(node)
            * depending upon the status of the execution, either move to the success node or failure node.

        Args:
            dag (Graph): The directed acyclic graph to traverse and execute.
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of the iterable.
                    Defaults to None.
        """
        ...

    @abstractmethod
    def _is_step_eligible_for_rerun(self, node: BaseNode, map_variable: TypeMapVariable = None):
        """
        In case of a re-run, this method checks to see if the previous run step status to determine if a re-run is
        necessary.
            * True: If its not a re-run.
            * True: If its a re-run and we failed in the last run or the corresponding logs do not exist.
            * False: If its a re-run and we succeeded in the last run.

        Most cases, this logic need not be touched

        Args:
            node (Node): The node to check against re-run
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of iterable..
                        Defaults to None.

        Returns:
            bool: Eligibility for re-run. True means re-run, False means skip to the next step.
        """
        ...

    @abstractmethod
    def send_return_code(self, stage="traversal"):
        """
        Convenience function used by pipeline to send return code to the caller of the cli

        Raises:
            Exception: If the pipeline execution failed
        """
        ...

    @abstractmethod
    def _resolve_executor_config(self, node: BaseNode):
        """
        The overrides section can contain specific over-rides to an global executor config.
        To avoid too much clutter in the dag definition, we allow the configuration file to have overrides block.
        The nodes can over-ride the global config by referring to key in the overrides.

        For example:
        # configuration.yaml
        execution:
          type: cloud-implementation
          config:
            k1: v1
            k3: v3
            overrides:
              k2: v2 # Could be a mapping internally.

        # in pipeline definition.yaml
        dag:
          steps:
            step1:
              overrides:
                cloud-implementation:
                  k1: value_specific_to_node
                  k2:

        This method should resolve the node_config to {'k1': 'value_specific_to_node', 'k2': 'v2', 'k3': 'v3'}

        Args:
            node (BaseNode): The current node being processed.

        """
        ...

    @abstractmethod
    def execute_job(self, node: TaskNode):
        """
        Executor specific way of executing a job (python function or a notebook).

        Interactive executors should execute the job.
        Transpilers should write the instructions.

        Args:
            node (BaseNode): The job node to execute

        Raises:
            NotImplementedError: Executors should choose to extend this functionality or not.
        """
        ...

    @abstractmethod
    def fan_out(self, node: BaseNode, map_variable: TypeMapVariable = None):
        """
        This method is used to appropriately fan-out the execution of a composite node.
        This is only useful when we want to execute a composite node during 3rd party orchestrators.

        Reason: Transpilers typically try to run the leaf nodes but do not have any capacity to do anything for the
        step which is composite. By calling this fan-out before calling the leaf nodes, we have an opportunity to
        do the right set up (creating the step log, exposing the parameters, etc.) for the composite step.

        All 3rd party orchestrators should use this method to fan-out the execution of a composite node.
        This ensures:
            - The dot path notation is preserved, this method should create the step and call the node's fan out to
            create the branch logs and let the 3rd party do the actual step execution.
            - Gives 3rd party orchestrators an opportunity to set out the required for running a composite node.

        Args:
            node (BaseNode): The node to fan-out
            map_variable (dict, optional): If the node if of a map state,.Defaults to None.

        """
        ...

    @abstractmethod
    def fan_in(self, node: BaseNode, map_variable: TypeMapVariable = None):
        """
        This method is used to appropriately fan-in after the execution of a composite node.
        This is only useful when we want to execute a composite node during 3rd party orchestrators.

        Reason: Transpilers typically try to run the leaf nodes but do not have any capacity to do anything for the
        step which is composite. By calling this fan-in after calling the leaf nodes, we have an opportunity to
        act depending upon the status of the individual branches.

        All 3rd party orchestrators should use this method to fan-in the execution of a composite node.
        This ensures:
            - Gives the renderer's the control on where to go depending upon the state of the composite node.
            - The status of the step and its underlying branches are correctly updated.

        Args:
            node (BaseNode): The node to fan-in
            map_variable (dict, optional): If the node if of a map state,.Defaults to None.

        """
        ...



import copy
import logging
from typing import Any, Dict, Type, cast

from pydantic import ConfigDict, Field

from magnus import context, defaults
from magnus.defaults import TypeMapVariable
from magnus.extensions.executor import GenericExecutor
from magnus.extensions.nodes import TaskNode
from magnus.integration import BaseIntegration
from magnus.nodes import BaseNode
from magnus.tasks import BaseTaskType

logger = logging.getLogger(defaults.LOGGER_NAME)


def create_executable(params: Dict[str, Any], model: Type[BaseTaskType], node_name: str) -> BaseTaskType:
    class EasyModel(model):  # type: ignore
        model_config = ConfigDict(extra="ignore")

    swallow_all = EasyModel(**params, node_name=node_name)
    return swallow_all


class MockedExecutor(GenericExecutor):
    service_name: str = "mocked"

    enable_parallel: bool = defaults.ENABLE_PARALLEL

    patches: Dict[str, Any] = Field(default_factory=dict)

    @property
    def _context(self):
        return context.run_context

    def _set_up_for_re_run(self, parameters: Dict[str, Any]) -> None:
        raise Exception("MockedExecutor does not support re-run")

    def execute_from_graph(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        This is the entry point to from the graph execution.

        While the self.execute_graph is responsible for traversing the graph, this function is responsible for
        actual execution of the node.

        If the node type is:
            * task : We can delegate to _execute_node after checking the eligibility for re-run in cases of a re-run
            * success: We can delegate to _execute_node
            * fail: We can delegate to _execute_node

        For nodes that are internally graphs:
            * parallel: Delegate the responsibility of execution to the node.execute_as_graph()
            * dag: Delegate the responsibility of execution to the node.execute_as_graph()
            * map: Delegate the responsibility of execution to the node.execute_as_graph()

        Transpilers will NEVER use this method and will NEVER call ths method.
        This method should only be used by interactive executors.

        Args:
            node (Node): The node to execute
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of iterable.
                    Defaults to None.
        """
        step_log = self._context.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable))

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING

        # Add the step log to the database as per the situation.
        # If its a terminal node, complete it now
        if node.node_type in ["success", "fail"]:
            self._context.run_log_store.add_step_log(step_log, self._context.run_id)
            self._execute_node(node, map_variable=map_variable, **kwargs)
            return

        # We call an internal function to iterate the sub graphs and execute them
        if node.is_composite:
            self._context.run_log_store.add_step_log(step_log, self._context.run_id)
            node.execute_as_graph(map_variable=map_variable, **kwargs)
            return

        node_to_send: TaskNode = cast(TaskNode, node).model_copy(deep=True)
        if node.name not in self.patches:
            # node is not patched, so mock it
            step_log.mock = True
        else:
            # node is mocked, change the executable to python with the
            # command as the patch value
            executable_type = node_to_send.executable.__class__
            executable = create_executable(
                self.patches[node.name],
                executable_type,
                node_name=node.name,
            )
            node_to_send.executable = executable
            pass

        # Executor specific way to trigger a job
        self._context.run_log_store.add_step_log(step_log, self._context.run_id)
        self.trigger_job(node=node_to_send, map_variable=map_variable, **kwargs)

    def trigger_job(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        Call this method only if we are responsible for traversing the graph via
        execute_from_graph().

        We are not prepared to execute node as of now.

        Args:
            node (BaseNode): The node to execute
            map_variable (str, optional): If the node if of a map state, this corresponds to the value of iterable.
                    Defaults to ''.

        NOTE: We do not raise an exception as this method is not required by many extensions
        """
        self.prepare_for_node_execution()
        self.execute_node(node=node, map_variable=map_variable, **kwargs)

    def _is_step_eligible_for_rerun(self, node: BaseNode, map_variable: TypeMapVariable = None):
        """
        In case of a re-run, this method checks to see if the previous run step status to determine if a re-run is
        necessary.
            * True: If its not a re-run.
            * True: If its a re-run and we failed in the last run or the corresponding logs do not exist.
            * False: If its a re-run and we succeeded in the last run.

        Most cases, this logic need not be touched

        Args:
            node (Node): The node to check against re-run
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of iterable..
                        Defaults to None.

        Returns:
            bool: Eligibility for re-run. True means re-run, False means skip to the next step.
        """
        return True

    def _resolve_executor_config(self, node: BaseNode):
        """
        The overrides section can contain specific over-rides to an global executor config.
        To avoid too much clutter in the dag definition, we allow the configuration file to have overrides block.
        The nodes can over-ride the global config by referring to key in the overrides.

        This function also applies variables to the effective node config.

        For example:
        # configuration.yaml
        execution:
          type: cloud-implementation
          config:
            k1: v1
            k3: v3
            overrides:
             custom_config:
                k1: v11
                k2: v2 # Could be a mapping internally.

        # in pipeline definition.yaml
        dag:
          steps:
            step1:
              overrides:
                cloud-implementation: custom_config

        This method should resolve the node_config to {'k1': 'v11', 'k2': 'v2', 'k3': 'v3'}

        Args:
            node (BaseNode): The current node being processed.

        """
        effective_node_config = copy.deepcopy(self.model_dump())

        return effective_node_config

    def execute_job(self, node: TaskNode):
        pass

    def execute_node(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        For local execution, we just execute the node.

        Args:
            node (BaseNode): _description_
            map_variable (dict[str, str], optional): _description_. Defaults to None.
        """
        self._execute_node(node=node, map_variable=map_variable, **kwargs)


class LocalContainerComputeFileSystemRunLogstore(BaseIntegration):
    """
    Integration between local container and file system run log store
    """

    executor_type = "local-container"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "file-system"  # The actual implementation of the service

    def validate(self, **kwargs):
        if self.executor._is_parallel_execution():  # pragma: no branch
            msg = "Mocked executor does not support parallel execution. "
            logger.warning(msg)


class LocalContainerComputeChunkedFSRunLogstore(BaseIntegration):
    """
    Integration between local container and file system run log store
    """

    executor_type = "local-container"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "chunked-fs"  # The actual implementation of the service

    def validate(self, **kwargs):
        if self.executor._is_parallel_execution():  # pragma: no branch
            msg = "Mocked executor does not support parallel execution. "
            logger.warning(msg)



from typing import Dict, Optional

from pydantic import BaseModel, SerializeAsAny

from magnus.catalog import BaseCatalog
from magnus.datastore import BaseRunLogStore
from magnus.executor import BaseExecutor
from magnus.experiment_tracker import BaseExperimentTracker
from magnus.graph import Graph
from magnus.secrets import BaseSecrets


class Context(BaseModel):
    executor: SerializeAsAny[BaseExecutor]
    run_log_store: SerializeAsAny[BaseRunLogStore]
    secrets_handler: SerializeAsAny[BaseSecrets]
    catalog_handler: SerializeAsAny[BaseCatalog]
    experiment_tracker: SerializeAsAny[BaseExperimentTracker]

    pipeline_file: Optional[str] = ""
    parameters_file: Optional[str] = ""
    configuration_file: Optional[str] = ""

    tag: str = ""
    run_id: str = ""
    variables: Dict[str, str] = {}
    use_cached: bool = False
    original_run_id: str = ""
    dag: Optional[Graph] = None
    dag_hash: str = ""
    execution_plan: str = ""


run_context = None  # type: Context # type: ignore



import logging
from pathlib import Path

from magnus import defaults
from magnus.extensions.run_log_store.file_system.implementation import FileSystemRunLogstore

logger = logging.getLogger(defaults.NAME)


class K8PersistentVolumeRunLogstore(FileSystemRunLogstore):
    """
    Uses the K8s Persistent Volumes to store run logs.
    """

    service_name: str = "k8s-pvc"
    persistent_volume_name: str
    mount_path: str

    @property
    def log_folder_name(self) -> str:
        return str(Path(self.mount_path) / self.log_folder)



import copy
import json
import logging
import os
from abc import abstractmethod
from typing import Any, Dict, List, Optional, cast

from rich import print

from magnus import context, defaults, exceptions, integration, parameters, utils
from magnus.datastore import DataCatalog, RunLog, StepLog
from magnus.defaults import TypeMapVariable
from magnus.executor import BaseExecutor
from magnus.experiment_tracker import get_tracked_data
from magnus.extensions.nodes import TaskNode
from magnus.graph import Graph
from magnus.nodes import BaseNode

logger = logging.getLogger(defaults.LOGGER_NAME)


class GenericExecutor(BaseExecutor):
    """
    The skeleton of an executor class.
    Any implementation of an executor should inherit this class and over-ride accordingly.

    This is a loaded base class which has a lot of methods already implemented for "typical" executions.
    Look at the function docs to understand how to use them appropriately.

    For any implementation:
    1). Who/when should the run log be set up?
    2). Who/When should the step log be set up?

    """

    service_name: str = ""
    service_type: str = "executor"

    @property
    def _context(self):
        return context.run_context

    @property
    def step_decorator_run_id(self):
        """
        TODO: Experimental feature, design is not mature yet.

        This function is used by the decorator function.
        The design idea is we can over-ride this method in different implementations to retrieve the run_id.
        But is it really intrusive to ask to set the environmental variable MAGNUS_RUN_ID?

        Returns:
            _type_: _description_
        """
        return os.environ.get("MAGNUS_RUN_ID", None)

    def _get_parameters(self) -> Dict[str, Any]:
        """
        Consolidate the parameters from the environment variables
        and the parameters file.

        The parameters defined in the environment variables take precedence over the parameters file.

        Returns:
            _type_: _description_
        """
        params: Dict[str, Any] = {}
        if self._context.parameters_file:
            params.update(utils.load_yaml(self._context.parameters_file))

        # Update these with some from the environment variables
        params.update(parameters.get_user_set_parameters())
        return params

    def _set_up_for_re_run(self, parameters: Dict[str, Any]) -> None:
        try:
            attempt_run_log = self._context.run_log_store.get_run_log_by_id(
                run_id=self._context.original_run_id, full=False
            )
        except exceptions.RunLogNotFoundError as e:
            msg = (
                f"Expected a run log with id: {self._context.original_run_id} "
                "but it does not exist in the run log store. "
                "If the original execution was in a different environment, ensure that it is available in the current "
                "environment."
            )
            logger.exception(msg)
            raise Exception(msg) from e

        # Sync the previous run log catalog to this one.
        self._context.catalog_handler.sync_between_runs(
            previous_run_id=self._context.original_run_id, run_id=self._context.run_id
        )

        parameters.update(cast(RunLog, attempt_run_log).parameters)

    def _set_up_run_log(self, exists_ok=False):
        """
        Create a run log and put that in the run log store

        If exists_ok, we allow the run log to be already present in the run log store.
        """
        try:
            attempt_run_log = self._context.run_log_store.get_run_log_by_id(run_id=self._context.run_id, full=False)

            logger.warning(f"The run log by id: {self._context.run_id} already exists")
            raise exceptions.RunLogExistsError(
                f"The run log by id: {self._context.run_id} already exists and is {attempt_run_log.status}"
            )
        except exceptions.RunLogNotFoundError:
            pass
        except exceptions.RunLogExistsError:
            if exists_ok:
                return
            raise

        # Consolidate and get the parameters
        parameters = self._get_parameters()

        if self._context.use_cached:
            self._set_up_for_re_run(parameters=parameters)

        self._context.run_log_store.create_run_log(
            run_id=self._context.run_id,
            tag=self._context.tag,
            status=defaults.PROCESSING,
            dag_hash=self._context.dag_hash,
            use_cached=self._context.use_cached,
            original_run_id=self._context.original_run_id,
        )
        # Any interaction with run log store attributes should happen via API if available.
        self._context.run_log_store.set_parameters(run_id=self._context.run_id, parameters=parameters)

        # Update run_config
        run_config = utils.get_run_config()
        self._context.run_log_store.set_run_config(run_id=self._context.run_id, run_config=run_config)

    def prepare_for_graph_execution(self):
        """
        This method should be called prior to calling execute_graph.
        Perform any steps required before doing the graph execution.

        The most common implementation is to prepare a run log for the run if the run uses local interactive compute.

        But in cases of actual rendering the job specs (eg: AWS step functions, K8's) we check if the services are OK.
        We do not set up a run log as its not relevant.
        """

        integration.validate(self, self._context.run_log_store)
        integration.configure_for_traversal(self, self._context.run_log_store)

        integration.validate(self, self._context.catalog_handler)
        integration.configure_for_traversal(self, self._context.catalog_handler)

        integration.validate(self, self._context.secrets_handler)
        integration.configure_for_traversal(self, self._context.secrets_handler)

        integration.validate(self, self._context.experiment_tracker)
        integration.configure_for_traversal(self, self._context.experiment_tracker)

        self._set_up_run_log()

    def prepare_for_node_execution(self):
        """
        Perform any modifications to the services prior to execution of the node.

        Args:
            node (Node): [description]
            map_variable (dict, optional): [description]. Defaults to None.
        """
        integration.validate(self, self._context.run_log_store)
        integration.configure_for_execution(self, self._context.run_log_store)

        integration.validate(self, self._context.catalog_handler)
        integration.configure_for_execution(self, self._context.catalog_handler)

        integration.validate(self, self._context.secrets_handler)
        integration.configure_for_execution(self, self._context.secrets_handler)

        integration.validate(self, self._context.experiment_tracker)
        integration.configure_for_execution(self, self._context.experiment_tracker)

    def _sync_catalog(self, step_log: StepLog, stage: str, synced_catalogs=None) -> Optional[List[DataCatalog]]:
        """
        1). Identify the catalog settings by over-riding node settings with the global settings.
        2). For stage = get:
                Identify the catalog items that are being asked to get from the catalog
                And copy them to the local compute data folder
        3). For stage = put:
                Identify the catalog items that are being asked to put into the catalog
                Copy the items from local compute folder to the catalog
        4). Add the items onto the step log according to the stage

        Args:
            node (Node): The current node being processed
            step_log (StepLog): The step log corresponding to that node
            stage (str): One of get or put

        Raises:
            Exception: If the stage is not in one of get/put

        """
        if stage not in ["get", "put"]:
            msg = (
                "Catalog service only accepts get/put possible actions as part of node execution."
                f"Sync catalog of the executor: {self.service_name} asks for {stage} which is not accepted"
            )
            raise Exception(msg)

        try:
            node_catalog_settings = self._context_node._get_catalog_settings()
        except exceptions.TerminalNodeError:
            return None

        if not (node_catalog_settings and stage in node_catalog_settings):
            logger.info("No catalog settings found for stage: %s", stage)
            # Nothing to get/put from the catalog
            return None

        compute_data_folder = self.get_effective_compute_data_folder()

        data_catalogs = []
        for name_pattern in node_catalog_settings.get(stage) or []:
            if stage == "get":
                data_catalog = self._context.catalog_handler.get(
                    name=name_pattern, run_id=self._context.run_id, compute_data_folder=compute_data_folder
                )
            elif stage == "put":
                data_catalog = self._context.catalog_handler.put(
                    name=name_pattern,
                    run_id=self._context.run_id,
                    compute_data_folder=compute_data_folder,
                    synced_catalogs=synced_catalogs,
                )
            else:
                raise Exception(f"Invalid stage: {stage}")
            logger.info(f"Added data catalog: {data_catalog} to step log")
            data_catalogs.extend(data_catalog)

        if data_catalogs:
            step_log.add_data_catalogs(data_catalogs)

        return data_catalogs

    def get_effective_compute_data_folder(self) -> str:
        """
        Get the effective compute data folder for the given stage.
        If there is nothing to catalog, we return None.

        The default is the compute data folder of the catalog but this can be over-ridden by the node.

        Args:
            stage (str): The stage we are in the process of cataloging


        Returns:
            str: The compute data folder as defined by the node defaulting to catalog handler
        """
        compute_data_folder = self._context.catalog_handler.compute_data_folder

        catalog_settings = self._context_node._get_catalog_settings()
        effective_compute_data_folder = catalog_settings.get("compute_data_folder", "") or compute_data_folder

        return effective_compute_data_folder

    @property
    def step_attempt_number(self) -> int:
        """
        The attempt number of the current step.
        Orchestrators should use this step to submit multiple attempts of the job.

        Returns:
            int: The attempt number of the current step. Defaults to 1.
        """
        return int(os.environ.get(defaults.ATTEMPT_NUMBER, 1))

    def _execute_node(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        This is the entry point when we do the actual execution of the function.
        DO NOT Over-ride this function.

        While in interactive execution, we just compute, in 3rd party interactive execution, we need to reach
        this function.

        In most cases,
            * We get the corresponding step_log of the node and the parameters.
            * We sync the catalog to GET any data sets that are in the catalog
            * We call the execute method of the node for the actual compute and retry it as many times as asked.
            * If the node succeeds, we get any of the user defined metrics provided by the user.
            * We sync the catalog to PUT any data sets that are in the catalog.

        Args:
            node (Node): The node to execute
            map_variable (dict, optional): If the node is of a map state, map_variable is the value of the iterable.
                        Defaults to None.
        """
        step_log = self._context.run_log_store.get_step_log(node._get_step_log_name(map_variable), self._context.run_id)
        """
        By now, all the parameters are part of the run log as a dictionary.
        We set them as environment variables, serialized as json strings.
        """
        params = self._context.run_log_store.get_parameters(run_id=self._context.run_id)
        parameters.set_user_defined_params_as_environment_variables(params)

        attempt = self.step_attempt_number
        logger.info(f"Trying to execute node: {node.internal_name}, attempt : {attempt}")

        attempt_log = self._context.run_log_store.create_attempt_log()
        self._context_step_log = step_log
        self._context_node = node

        data_catalogs_get: Optional[List[DataCatalog]] = self._sync_catalog(step_log, stage="get")
        try:
            attempt_log = node.execute(executor=self, mock=step_log.mock, map_variable=map_variable, **kwargs)
        except Exception as e:
            # Any exception here is a magnus exception as node suppresses exceptions.
            msg = "This is clearly magnus fault, please report a bug and the logs"
            logger.exception(msg)
            raise Exception(msg) from e
        finally:
            attempt_log.attempt_number = attempt
            attempt_log.parameters = params.copy()
            step_log.attempts.append(attempt_log)

            tracked_data = get_tracked_data()

            self._context.experiment_tracker.publish_data(tracked_data)
            # By this point, the updated parameters are deserialized as json strings.
            parameters_out = parameters.get_user_set_parameters(remove=True)

            if attempt_log.status == defaults.FAIL:
                logger.exception(f"Node: {node} failed")
                step_log.status = defaults.FAIL
            else:
                # Mock is always set to False, bad design??
                # TODO: Stub nodes should not sync back data
                # TODO: Errors in catalog syncing should point to Fail step
                # TODO: Even for a failed execution, the catalog can happen
                step_log.status = defaults.SUCCESS
                self._sync_catalog(step_log, stage="put", synced_catalogs=data_catalogs_get)
                step_log.user_defined_metrics = tracked_data
                diff_parameters = utils.diff_dict(params, parameters_out)
                self._context.run_log_store.set_parameters(self._context.run_id, diff_parameters)

            # Remove the step context
            self._context_step_log = None
            self._context_node = None  # type: ignore
            self._context_metrics = {}  # type: ignore

            self._context.run_log_store.add_step_log(step_log, self._context.run_id)

    def add_code_identities(self, node: BaseNode, step_log: StepLog, **kwargs):
        """
        Add code identities specific to the implementation.

        The Base class has an implementation of adding git code identities.

        Args:
            step_log (object): The step log object
            node (BaseNode): The node we are adding the step log for
        """
        step_log.code_identities.append(utils.get_git_code_identity())

    def execute_from_graph(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        This is the entry point to from the graph execution.

        While the self.execute_graph is responsible for traversing the graph, this function is responsible for
        actual execution of the node.

        If the node type is:
            * task : We can delegate to _execute_node after checking the eligibility for re-run in cases of a re-run
            * success: We can delegate to _execute_node
            * fail: We can delegate to _execute_node

        For nodes that are internally graphs:
            * parallel: Delegate the responsibility of execution to the node.execute_as_graph()
            * dag: Delegate the responsibility of execution to the node.execute_as_graph()
            * map: Delegate the responsibility of execution to the node.execute_as_graph()

        Transpilers will NEVER use this method and will NEVER call ths method.
        This method should only be used by interactive executors.

        Args:
            node (Node): The node to execute
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of iterable.
                    Defaults to None.
        """
        step_log = self._context.run_log_store.create_step_log(node.name, node._get_step_log_name(map_variable))

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING

        # Add the step log to the database as per the situation.
        # If its a terminal node, complete it now
        if node.node_type in ["success", "fail"]:
            self._context.run_log_store.add_step_log(step_log, self._context.run_id)
            self._execute_node(node, map_variable=map_variable, **kwargs)
            return

        # In single step
        if (self._single_step and not node.name == self._single_step) or not self._is_step_eligible_for_rerun(
            node, map_variable=map_variable
        ):
            # If the node name does not match, we move on to the next node.
            # If previous run was successful, move on to the next step
            step_log.mock = True
            step_log.status = defaults.SUCCESS
            self._context.run_log_store.add_step_log(step_log, self._context.run_id)
            return
        # We call an internal function to iterate the sub graphs and execute them
        if node.is_composite:
            self._context.run_log_store.add_step_log(step_log, self._context.run_id)
            node.execute_as_graph(map_variable=map_variable, **kwargs)
            return

        # Executor specific way to trigger a job
        self._context.run_log_store.add_step_log(step_log, self._context.run_id)
        self.trigger_job(node=node, map_variable=map_variable, **kwargs)

    def trigger_job(self, node: BaseNode, map_variable: TypeMapVariable = None, **kwargs):
        """
        Call this method only if we are responsible for traversing the graph via
        execute_from_graph().

        We are not prepared to execute node as of now.

        Args:
            node (BaseNode): The node to execute
            map_variable (str, optional): If the node if of a map state, this corresponds to the value of iterable.
                    Defaults to ''.

        NOTE: We do not raise an exception as this method is not required by many extensions
        """
        pass

    def _get_status_and_next_node_name(self, current_node: BaseNode, dag: Graph, map_variable: TypeMapVariable = None):
        """
        Given the current node and the graph, returns the name of the next node to execute.

        The name is always relative the graph that the node resides in.

        If the current node succeeded, we return the next node as per the graph.
        If the current node failed, we return the on failure node of the node (if provided) or the global one.

        This method is only used by interactive executors i.e local and local-container

        Args:
            current_node (BaseNode): The current node.
            dag (Graph): The dag we are traversing.
            map_variable (dict): If the node belongs to a map branch.

        """

        step_log = self._context.run_log_store.get_step_log(
            current_node._get_step_log_name(map_variable), self._context.run_id
        )
        logger.info(f"Finished executing the node {current_node} with status {step_log.status}")

        try:
            next_node_name = current_node._get_next_node()
        except exceptions.TerminalNodeError:
            next_node_name = ""

        if step_log.status == defaults.FAIL:
            next_node_name = dag.get_fail_node().name
            if current_node._get_on_failure_node():
                next_node_name = current_node._get_on_failure_node()

        return step_log.status, next_node_name

    def execute_graph(self, dag: Graph, map_variable: TypeMapVariable = None, **kwargs):
        """
        The parallelization is controlled by the nodes and not by this function.

        Transpilers should over ride this method to do the translation of dag to the platform specific way.
        Interactive methods should use this to traverse and execute the dag.
            - Use execute_from_graph to handle sub-graphs

        Logically the method should:
            * Start at the dag.start_at of the dag.
            * Call the self.execute_from_graph(node)
            * depending upon the status of the execution, either move to the success node or failure node.

        Args:
            dag (Graph): The directed acyclic graph to traverse and execute.
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of the iterable.
                    Defaults to None.
        """
        current_node = dag.start_at
        previous_node = None
        logger.info(f"Running the execution with {current_node}")

        while True:
            working_on = dag.get_node_by_name(current_node)

            if previous_node == current_node:
                raise Exception("Potentially running in a infinite loop")

            previous_node = current_node

            logger.info(f"Creating execution log for {working_on}")
            self.execute_from_graph(working_on, map_variable=map_variable, **kwargs)

            status, next_node_name = self._get_status_and_next_node_name(
                current_node=working_on, dag=dag, map_variable=map_variable
            )

            if status == defaults.TRIGGERED:
                # Some nodes go into triggered state and self traverse
                logger.info(f"Triggered the job to execute the node {current_node}")
                break

            if working_on.node_type in ["success", "fail"]:
                break

            current_node = next_node_name

        run_log = self._context.run_log_store.get_branch_log(
            working_on._get_branch_log_name(map_variable), self._context.run_id
        )

        branch = "graph"
        if working_on.internal_branch_name:
            branch = working_on.internal_branch_name

        logger.info(f"Finished execution of the {branch} with status {run_log.status}")

        # get the final run log
        if branch == "graph":
            run_log = self._context.run_log_store.get_run_log_by_id(run_id=self._context.run_id, full=True)
        print(json.dumps(run_log.model_dump(), indent=4))

    def _is_step_eligible_for_rerun(self, node: BaseNode, map_variable: TypeMapVariable = None):
        """
        In case of a re-run, this method checks to see if the previous run step status to determine if a re-run is
        necessary.
            * True: If its not a re-run.
            * True: If its a re-run and we failed in the last run or the corresponding logs do not exist.
            * False: If its a re-run and we succeeded in the last run.

        Most cases, this logic need not be touched

        Args:
            node (Node): The node to check against re-run
            map_variable (dict, optional): If the node if of a map state, this corresponds to the value of iterable..
                        Defaults to None.

        Returns:
            bool: Eligibility for re-run. True means re-run, False means skip to the next step.
        """
        if self._context.use_cached:
            node_step_log_name = node._get_step_log_name(map_variable=map_variable)
            logger.info(f"Scanning previous run logs for node logs of: {node_step_log_name}")

            try:
                previous_node_log = self._context.run_log_store.get_step_log(
                    internal_name=node_step_log_name, run_id=self._context.original_run_id
                )
            except exceptions.StepLogNotFoundError:
                logger.warning(f"Did not find the node {node.name} in previous run log")
                return True  # We should re-run the node.

            logger.info(f"The original step status: {previous_node_log.status}")

            if previous_node_log.status == defaults.SUCCESS:
                return False  # We need not run the node

            logger.info(f"The new execution should start executing graph from this node {node.name}")
            return True

        return True

    def send_return_code(self, stage="traversal"):
        """
        Convenience function used by pipeline to send return code to the caller of the cli

        Raises:
            Exception: If the pipeline execution failed
        """
        run_id = self._context.run_id

        run_log = self._context.run_log_store.get_run_log_by_id(run_id=run_id, full=False)
        if run_log.status == defaults.FAIL:
            raise exceptions.ExecutionFailedError(run_id=run_id)

    def _resolve_executor_config(self, node: BaseNode):
        """
        The overrides section can contain specific over-rides to an global executor config.
        To avoid too much clutter in the dag definition, we allow the configuration file to have overrides block.
        The nodes can over-ride the global config by referring to key in the overrides.

        This function also applies variables to the effective node config.

        For example:
        # configuration.yaml
        execution:
          type: cloud-implementation
          config:
            k1: v1
            k3: v3
            overrides:
             custom_config:
                k1: v11
                k2: v2 # Could be a mapping internally.

        # in pipeline definition.yaml
        dag:
          steps:
            step1:
              overrides:
                cloud-implementation: custom_config

        This method should resolve the node_config to {'k1': 'v11', 'k2': 'v2', 'k3': 'v3'}

        Args:
            node (BaseNode): The current node being processed.

        """
        effective_node_config = copy.deepcopy(self.model_dump())
        try:
            ctx_node_config = node._get_executor_config(self.service_name)
        except exceptions.TerminalNodeError:
            # Some modes request for effective node config even for success or fail nodes
            return utils.apply_variables(effective_node_config, self._context.variables)

        if ctx_node_config:
            if ctx_node_config not in self.overrides:
                raise Exception(f"No override of key: {ctx_node_config} found in the overrides section")

            effective_node_config.update(self.overrides[ctx_node_config])

        effective_node_config = utils.apply_variables(effective_node_config, self._context.variables)
        logger.debug(f"Effective node config: {effective_node_config}")

        return effective_node_config

    @abstractmethod
    def execute_job(self, node: TaskNode):
        """
        Executor specific way of executing a job (python function or a notebook).

        Interactive executors should execute the job.
        Transpilers should write the instructions.

        Args:
            node (BaseNode): The job node to execute

        Raises:
            NotImplementedError: Executors should choose to extend this functionality or not.
        """
        raise NotImplementedError

    def fan_out(self, node: BaseNode, map_variable: TypeMapVariable = None):
        """
        This method is used to appropriately fan-out the execution of a composite node.
        This is only useful when we want to execute a composite node during 3rd party orchestrators.

        Reason: Transpilers typically try to run the leaf nodes but do not have any capacity to do anything for the
        step which is composite. By calling this fan-out before calling the leaf nodes, we have an opportunity to
        do the right set up (creating the step log, exposing the parameters, etc.) for the composite step.

        All 3rd party orchestrators should use this method to fan-out the execution of a composite node.
        This ensures:
            - The dot path notation is preserved, this method should create the step and call the node's fan out to
            create the branch logs and let the 3rd party do the actual step execution.
            - Gives 3rd party orchestrators an opportunity to set out the required for running a composite node.

        Args:
            node (BaseNode): The node to fan-out
            map_variable (dict, optional): If the node if of a map state,.Defaults to None.

        """
        step_log = self._context.run_log_store.create_step_log(
            node.name, node._get_step_log_name(map_variable=map_variable)
        )

        self.add_code_identities(node=node, step_log=step_log)

        step_log.step_type = node.node_type
        step_log.status = defaults.PROCESSING
        self._context.run_log_store.add_step_log(step_log, self._context.run_id)

        node.fan_out(executor=self, map_variable=map_variable)

    def fan_in(self, node: BaseNode, map_variable: TypeMapVariable = None):
        """
        This method is used to appropriately fan-in after the execution of a composite node.
        This is only useful when we want to execute a composite node during 3rd party orchestrators.

        Reason: Transpilers typically try to run the leaf nodes but do not have any capacity to do anything for the
        step which is composite. By calling this fan-in after calling the leaf nodes, we have an opportunity to
        act depending upon the status of the individual branches.

        All 3rd party orchestrators should use this method to fan-in the execution of a composite node.
        This ensures:
            - Gives the renderer's the control on where to go depending upon the state of the composite node.
            - The status of the step and its underlying branches are correctly updated.

        Args:
            node (BaseNode): The node to fan-in
            map_variable (dict, optional): If the node if of a map state,.Defaults to None.

        """
        node.fan_in(executor=self, map_variable=map_variable)

        step_log = self._context.run_log_store.get_step_log(
            node._get_step_log_name(map_variable=map_variable), self._context.run_id
        )

        if step_log.status == defaults.FAIL:
            raise Exception(f"Step {node.name} failed")



import logging
from typing import cast

from magnus import defaults
from magnus.integration import BaseIntegration

logger = logging.getLogger(defaults.NAME)


class LocalCompute(BaseIntegration):
    """
    Integration between local and k8's pvc
    """

    executor_type = "local"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "chunked-k8s-pvc"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = "We can't use the local compute k8s pvc store integration."
        raise Exception(msg)


class LocalContainerCompute(BaseIntegration):
    """
    Integration between local-container and k8's pvc
    """

    executor_type = "local-container"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "chunked-k8s-pvc"  # The actual implementation of the service

    def validate(self, **kwargs):
        msg = "We can't use the local-container compute k8s pvc store integration."
        raise Exception(msg)


class ArgoCompute(BaseIntegration):
    """
    Integration between argo and k8's pvc
    """

    executor_type = "argo"
    service_type = "run_log_store"  # One of secret, catalog, datastore
    service_provider = "chunked-k8s-pvc"  # The actual implementation of the service

    def configure_for_traversal(self, **kwargs):
        from magnus.extensions.executor.argo.implementation import ArgoExecutor, UserVolumeMounts
        from magnus.extensions.run_log_store.chunked_k8s_pvc.implementation import ChunkedK8PersistentVolumeRunLogstore

        self.executor = cast(ArgoExecutor, self.executor)
        self.service = cast(ChunkedK8PersistentVolumeRunLogstore, self.service)

        volume_mount = UserVolumeMounts(
            name=self.service.persistent_volume_name,
            mount_path=self.service.mount_path,
        )

        self.executor.persistent_volumes.append(volume_mount)



import logging

import click
from click_plugins import with_plugins
from pkg_resources import iter_entry_points

from magnus import defaults, entrypoints

logger = logging.getLogger(defaults.LOGGER_NAME)


@with_plugins(iter_entry_points("magnus.cli_plugins"))
@click.group()
@click.version_option()
def cli():
    """
    Welcome to magnus. Please provide the command that you want to use.
    All commands have options that you can see by magnus <command> --help
    """
    pass


@cli.command("execute", short_help="Execute/translate a pipeline")
@click.option("-f", "--file", default="pipeline.yaml", help="The pipeline definition file", show_default=True)
@click.option(
    "-c", "--config-file", default=None, help="config file, in yaml, to be used for the run", show_default=True
)
@click.option(
    "-p",
    "--parameters-file",
    default=None,
    help="Parameters, in yaml,  accessible by the application",
    show_default=True,
)
@click.option(
    "--log-level",
    default=defaults.LOG_LEVEL,
    help="The log level",
    show_default=True,
    type=click.Choice(["INFO", "DEBUG", "WARNING", "ERROR", "FATAL"]),
)
@click.option("--tag", default="", help="A tag attached to the run")
@click.option("--run-id", help="An optional run_id, one would be generated if not provided")
@click.option("--use-cached", help="Provide the previous run_id to re-run.", show_default=True)
def execute(file, config_file, parameters_file, log_level, tag, run_id, use_cached):  # pragma: no cover
    """
    Execute a pipeline

    Usage: magnus execute [OPTIONS]

    Options:
    -f, --file TEXT               The pipeline definition file [default: pipeline.yaml]
    -c, --config-file TEXT        config file, in yaml, to be used for the run [default: None]
    -p, --parameters-file TEXT    Parameters, in yaml,  accessible by the application [default: None]
    --log-level                     One of [INFO|DEBUG|WARNING|ERROR|FATAL]
                                    The log level
                                    [default: INFO]
    --tag TEXT                   A tag attached to the run
                                    [default: ]
    --run-id TEXT                An optional run_id, one would be generated if not
                                    provided
    --use-cached TEXT            Provide the previous run_id to re-run.
    """
    logger.setLevel(log_level)
    entrypoints.execute(
        configuration_file=config_file,
        pipeline_file=file,
        tag=tag,
        run_id=run_id,
        use_cached=use_cached,
        parameters_file=parameters_file,
    )


@cli.command("execute_single_node", short_help="Internal entry point to execute a single node", hidden=True)
@click.argument("run_id")
@click.argument("step_name")
@click.option("--map-variable", default="", help="The map variable dictionary in str", show_default=True)
@click.option("-f", "--file", default="", help="The pipeline definition file", show_default=True)
@click.option(
    "-c", "--config-file", default=None, help="config file, in yaml, to be used for the run", show_default=True
)
@click.option(
    "-p",
    "--parameters-file",
    default=None,
    help="Parameters, in yaml,  accessible by the application",
    show_default=True,
)
@click.option(
    "--log-level",
    default=defaults.LOG_LEVEL,
    help="The log level",
    show_default=True,
    type=click.Choice(["INFO", "DEBUG", "WARNING", "ERROR", "FATAL"]),
)
@click.option("--tag", default="", help="A tag attached to the run")
def execute_single_node(run_id, step_name, map_variable, file, config_file, parameters_file, log_level, tag):
    """
    Internal entrypoint for magnus to execute a single node.

    Other than local executor, every other executor uses this entry point to execute a step in the context of magnus.
    Only chained executions should use this method. Unchained executions should use execute_
    """
    logger.setLevel(log_level)

    # Execute the node as part of the graph execution.
    entrypoints.execute_single_node(
        configuration_file=config_file,
        pipeline_file=file,
        step_name=step_name,
        map_variable=map_variable,
        run_id=run_id,
        tag=tag,
        parameters_file=parameters_file,
    )


@cli.command("execute_notebook", short_help="Entry point to execute a notebook")
@click.argument("filename")
@click.option("--entrypoint", default=defaults.ENTRYPOINT.USER.value, hidden=True)
@click.option(
    "-c", "--config-file", default=None, help="config file, in yaml, to be used for the run", show_default=True
)
@click.option(
    "-p",
    "--parameters-file",
    default=None,
    help="Parameters, in yaml,  accessible by the application",
    show_default=True,
)
@click.option(
    "--log-level",
    default=defaults.LOG_LEVEL,
    help="The log level",
    show_default=True,
    type=click.Choice(["INFO", "DEBUG", "WARNING", "ERROR", "FATAL"]),
)
@click.option("--data-folder", "-d", default="data/", help="The catalog data folder")
@click.option("--put-in-catalog", "-put", default=None, multiple=True, help="The data to put from the catalog")
@click.option("--notebook-output-path", default="", help="The output path for the notebook")
@click.option("--tag", help="A tag attached to the run")
@click.option("--run-id", help="An optional run_id, one would be generated if not provided")
def execute_notebook(
    filename,
    entrypoint,
    config_file,
    parameters_file,
    log_level,
    data_folder,
    put_in_catalog,
    notebook_output_path,
    tag,
    run_id,
):
    """
    External entry point to execute a Jupyter notebook in isolation.

    The notebook would be executed in the environment defined by the config file or default if none.
    The execution plan is unchained.
    """
    logger.setLevel(log_level)
    catalog_config = {"compute_data_folder": data_folder, "put": list(put_in_catalog) if put_in_catalog else None}
    if not filename.endswith(".ipynb"):
        raise Exception("A notebook should always have ipynb as the extension")

    entrypoints.execute_notebook(
        entrypoint=entrypoint,
        notebook_file=filename,
        catalog_config=catalog_config,
        configuration_file=config_file,
        parameters_file=parameters_file,
        notebook_output_path=notebook_output_path,
        tag=tag,
        run_id=run_id,
    )


@cli.command("execute_function", short_help="Entry point to execute a python function")
@click.argument("command")
@click.option("--entrypoint", default=defaults.ENTRYPOINT.USER.value, hidden=True)
@click.option(
    "-c", "--config-file", default=None, help="config file, in yaml, to be used for the run", show_default=True
)
@click.option(
    "-p",
    "--parameters-file",
    default=None,
    help="Parameters, in yaml,  accessible by the application",
    show_default=True,
)
@click.option(
    "--log-level",
    default=defaults.LOG_LEVEL,
    help="The log level",
    show_default=True,
    type=click.Choice(["INFO", "DEBUG", "WARNING", "ERROR", "FATAL"]),
)
@click.option("--data-folder", "-d", default="data/", help="The catalog data folder")
@click.option("--put-in-catalog", "-put", default=None, multiple=True, help="The data to put from the catalog")
@click.option("--tag", help="A tag attached to the run")
@click.option("--run-id", help="An optional run_id, one would be generated if not provided")
def execute_function(
    command, entrypoint, config_file, parameters_file, log_level, data_folder, put_in_catalog, tag, run_id
):
    """
    External entry point to execute a python function in isolation.

    The function would be executed in the environment defined by the config file or default if none.
    The execution plan is unchained.
    """
    logger.setLevel(log_level)
    catalog_config = {"compute_data_folder": data_folder, "put": list(put_in_catalog) if put_in_catalog else None}
    entrypoints.execute_function(
        entrypoint=entrypoint,
        command=command,
        catalog_config=catalog_config,
        configuration_file=config_file,
        parameters_file=parameters_file,
        tag=tag,
        run_id=run_id,
    )


@cli.command("fan", short_help="Internal entry point to fan in or out a composite node", hidden=True)
@click.argument("run_id")
@click.argument("step_name")
@click.option("-m", "--mode", help="fan in or fan out", required=True, type=click.Choice(["in", "out"]))
@click.option("--map-variable", default="", help="The map variable dictionary in str", show_default=True)
@click.option("-f", "--file", default="", help="The pipeline definition file", show_default=True)
@click.option(
    "-c", "--config-file", default=None, help="config file, in yaml, to be used for the run", show_default=True
)
@click.option(
    "-p",
    "--parameters-file",
    default=None,
    help="Parameters, in yaml,  accessible by the application",
    show_default=True,
)
@click.option(
    "--log-level",
    default=defaults.LOG_LEVEL,
    help="The log level",
    show_default=True,
    type=click.Choice(["INFO", "DEBUG", "WARNING", "ERROR", "FATAL"]),
)
@click.option("--tag", default="", help="A tag attached to the run")
def fan(run_id, step_name, mode, map_variable, file, config_file, parameters_file, log_level, tag):
    """
    Internal entrypoint for magnus to fan in or out a composite node.

    Only 3rd party orchestrators should use this entry point.
    """
    logger.setLevel(log_level)

    # Fan in or out
    entrypoints.fan(
        configuration_file=config_file,
        pipeline_file=file,
        step_name=step_name,
        mode=mode,
        map_variable=map_variable,
        run_id=run_id,
        tag=tag,
        parameters_file=parameters_file,
    )


# Needed for the binary creation
if __name__ == "__main__":
    cli()



import json
import logging
import multiprocessing
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from typing import Any, Dict, cast

from pydantic import ConfigDict, Field, ValidationInfo, field_serializer, field_validator
from typing_extensions import Annotated

from magnus import defaults, utils
from magnus.datastore import StepAttempt
from magnus.defaults import TypeMapVariable
from magnus.graph import Graph, create_graph
from magnus.nodes import CompositeNode, ExecutableNode, TerminalNode
from magnus.tasks import BaseTaskType, create_task

logger = logging.getLogger(defaults.LOGGER_NAME)


class TaskNode(ExecutableNode):
    """
    A node of type Task.

    This node does the actual function execution of the graph in all cases.
    """

    executable: BaseTaskType = Field(exclude=True)
    node_type: str = Field(default="task", serialization_alias="type")

    # It is technically not allowed as parse_from_config filters them.
    # This is just to get the task level configuration to be present during serialization.
    model_config = ConfigDict(extra="allow")

    @classmethod
    def parse_from_config(cls, config: Dict[str, Any]) -> "TaskNode":
        # separate task config from node config
        task_config = {k: v for k, v in config.items() if k not in TaskNode.model_fields.keys()}
        node_config = {k: v for k, v in config.items() if k in TaskNode.model_fields.keys()}

        task_config["node_name"] = config.get("name")

        executable = create_task(task_config)
        return cls(executable=executable, **node_config, **task_config)

    def execute(self, mock=False, map_variable: TypeMapVariable = None, **kwargs) -> StepAttempt:
        """
        All that we do in magnus is to come to this point where we actually execute the command.

        Args:
            executor (_type_): The executor class
            mock (bool, optional): If we should just mock and not execute. Defaults to False.
            map_variable (dict, optional): If the node is part of internal branch. Defaults to None.

        Returns:
            StepAttempt: The attempt object
        """
        print("Executing task:", self._context.executor._context_node)
        # Here is where the juice is
        attempt_log = self._context.run_log_store.create_attempt_log()
        try:
            attempt_log.start_time = str(datetime.now())
            attempt_log.status = defaults.SUCCESS
            if not mock:
                # Do not run if we are mocking the execution, could be useful for caching and dry runs
                self.executable.execute_command(map_variable=map_variable)
        except Exception as _e:  # pylint: disable=W0703
            logger.exception("Task failed")
            attempt_log.status = defaults.FAIL
            attempt_log.message = str(_e)
        finally:
            attempt_log.end_time = str(datetime.now())
            attempt_log.duration = utils.get_duration_between_datetime_strings(
                attempt_log.start_time, attempt_log.end_time
            )
        return attempt_log


class FailNode(TerminalNode):
    """
    A leaf node of the graph that represents a failure node
    """

    node_type: str = Field(default="fail", serialization_alias="type")

    @classmethod
    def parse_from_config(cls, config: Dict[str, Any]) -> "FailNode":
        return cast("FailNode", super().parse_from_config(config))

    def execute(self, mock=False, map_variable: TypeMapVariable = None, **kwargs) -> StepAttempt:
        """
        Execute the failure node.
        Set the run or branch log status to failure.

        Args:
            executor (_type_): the executor class
            mock (bool, optional): If we should just mock and not do the actual execution. Defaults to False.
            map_variable (dict, optional): If the node belongs to internal branches. Defaults to None.

        Returns:
            StepAttempt: The step attempt object
        """
        attempt_log = self._context.run_log_store.create_attempt_log()
        try:
            attempt_log.start_time = str(datetime.now())
            attempt_log.status = defaults.SUCCESS
            #  could be a branch or run log
            run_or_branch_log = self._context.run_log_store.get_branch_log(
                self._get_branch_log_name(map_variable), self._context.run_id
            )
            run_or_branch_log.status = defaults.FAIL
            self._context.run_log_store.add_branch_log(run_or_branch_log, self._context.run_id)
        except BaseException:  # pylint: disable=W0703
            logger.exception("Fail node execution failed")
        finally:
            attempt_log.status = defaults.SUCCESS  # This is a dummy node, so we ignore errors and mark SUCCESS
            attempt_log.end_time = str(datetime.now())
            attempt_log.duration = utils.get_duration_between_datetime_strings(
                attempt_log.start_time, attempt_log.end_time
            )
        return attempt_log


class SuccessNode(TerminalNode):
    """
    A leaf node of the graph that represents a success node
    """

    node_type: str = Field(default="success", serialization_alias="type")

    @classmethod
    def parse_from_config(cls, config: Dict[str, Any]) -> "SuccessNode":
        return cast("SuccessNode", super().parse_from_config(config))

    def execute(self, mock=False, map_variable: TypeMapVariable = None, **kwargs) -> StepAttempt:
        """
        Execute the success node.
        Set the run or branch log status to success.

        Args:
            executor (_type_): The executor class
            mock (bool, optional): If we should just mock and not perform anything. Defaults to False.
            map_variable (dict, optional): If the node belongs to an internal branch. Defaults to None.

        Returns:
            StepAttempt: The step attempt object
        """
        attempt_log = self._context.run_log_store.create_attempt_log()
        try:
            attempt_log.start_time = str(datetime.now())
            attempt_log.status = defaults.SUCCESS
            #  could be a branch or run log
            run_or_branch_log = self._context.run_log_store.get_branch_log(
                self._get_branch_log_name(map_variable), self._context.run_id
            )
            run_or_branch_log.status = defaults.SUCCESS
            self._context.run_log_store.add_branch_log(run_or_branch_log, self._context.run_id)
        except BaseException:  # pylint: disable=W0703
            logger.exception("Success node execution failed")
        finally:
            attempt_log.status = defaults.SUCCESS  # This is a dummy node and we make sure we mark it as success
            attempt_log.end_time = str(datetime.now())
            attempt_log.duration = utils.get_duration_between_datetime_strings(
                attempt_log.start_time, attempt_log.end_time
            )
        return attempt_log


class ParallelNode(CompositeNode):
    """
    A composite node containing many graph objects within itself.

    The structure is generally:
        ParallelNode:
            Branch A:
                Sub graph definition
            Branch B:
                Sub graph definition
            . . .

    """

    node_type: str = Field(default="parallel", serialization_alias="type")
    branches: Dict[str, Graph]
    is_composite: bool = Field(default=True, exclude=True)

    @field_serializer("branches")
    def ser_branches(self, branches: Dict[str, Graph]) -> Dict[str, Graph]:
        ret: Dict[str, Graph] = {}

        for branch_name, branch in branches.items():
            ret[branch_name.split(".")[-1]] = branch

        return ret

    @classmethod
    def parse_from_config(cls, config: Dict[str, Any]) -> "ParallelNode":
        internal_name = cast(str, config.get("internal_name"))

        config_branches = config.pop("branches", {})
        branches = {}
        for branch_name, branch_config in config_branches.items():
            sub_graph = create_graph(
                deepcopy(branch_config),
                internal_branch_name=internal_name + "." + branch_name,
            )
            branches[internal_name + "." + branch_name] = sub_graph

        if not branches:
            raise Exception("A parallel node should have branches")
        return cls(branches=branches, **config)

    def _get_branch_by_name(self, branch_name: str) -> Graph:
        if branch_name in self.branches:
            return self.branches[branch_name]

        raise Exception(f"Branch {branch_name} does not exist")

    def fan_out(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        The general fan out method for a node of type Parallel.
        This method assumes that the step log has already been created.

        3rd party orchestrators should create the step log and use this method to create the branch logs.

        Args:
            executor (BaseExecutor): The executor class as defined by the config
            map_variable (dict, optional): If the node is part of a map node. Defaults to None.
        """
        # Prepare the branch logs
        for internal_branch_name, _ in self.branches.items():
            effective_branch_name = self._resolve_map_placeholders(internal_branch_name, map_variable=map_variable)

            branch_log = self._context.run_log_store.create_branch_log(effective_branch_name)
            branch_log.status = defaults.PROCESSING
            self._context.run_log_store.add_branch_log(branch_log, self._context.run_id)

    def execute_as_graph(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        This function does the actual execution of the sub-branches of the parallel node.

        From a design perspective, this function should not be called if the execution is 3rd party orchestrated.

        The modes that render the job specifications, do not need to interact with this node at all as they have their
        own internal mechanisms of handing parallel states.
        If they do not, you can find a way using as-is nodes as hack nodes.

        The execution of a dag, could result in
            * The dag being completely executed with a definite (fail, success) state in case of
                local or local-container execution
            * The dag being in a processing state with PROCESSING status in case of local-aws-batch

        Only fail state is considered failure during this phase of execution.

        Args:
            executor (Executor): The Executor as per the use config
            **kwargs: Optional kwargs passed around
        """
        from magnus import entrypoints

        self.fan_out(map_variable=map_variable, **kwargs)

        jobs = []
        # Given that we can have nesting and complex graphs, controlling the number of processes is hard.
        # A better way is to actually submit the job to some process scheduler which does resource management
        for internal_branch_name, branch in self.branches.items():
            if self._context.executor._is_parallel_execution():
                # Trigger parallel jobs
                action = entrypoints.execute_single_brach
                kwargs = {
                    "configuration_file": self._context.configuration_file,
                    "pipeline_file": self._context.pipeline_file,
                    "branch_name": internal_branch_name.replace(" ", defaults.COMMAND_FRIENDLY_CHARACTER),
                    "run_id": self._context.run_id,
                    "map_variable": json.dumps(map_variable),
                    "tag": self._context.tag,
                }
                process = multiprocessing.Process(target=action, kwargs=kwargs)
                jobs.append(process)
                process.start()

            else:
                # If parallel is not enabled, execute them sequentially
                self._context.executor.execute_graph(branch, map_variable=map_variable, **kwargs)

        for job in jobs:
            job.join()  # Find status of the branches

        self.fan_in(map_variable=map_variable, **kwargs)

    def fan_in(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        The general fan in method for a node of type Parallel.

        3rd party orchestrators should use this method to find the status of the composite step.

        Args:
            executor (BaseExecutor): The executor class as defined by the config
            map_variable (dict, optional): If the node is part of a map. Defaults to None.
        """
        step_success_bool = True
        for internal_branch_name, _ in self.branches.items():
            effective_branch_name = self._resolve_map_placeholders(internal_branch_name, map_variable=map_variable)
            branch_log = self._context.run_log_store.get_branch_log(effective_branch_name, self._context.run_id)
            if branch_log.status != defaults.SUCCESS:
                step_success_bool = False

        # Collate all the results and update the status of the step
        effective_internal_name = self._resolve_map_placeholders(self.internal_name, map_variable=map_variable)
        step_log = self._context.run_log_store.get_step_log(effective_internal_name, self._context.run_id)

        if step_success_bool:  #  If none failed
            step_log.status = defaults.SUCCESS
        else:
            step_log.status = defaults.FAIL

        self._context.run_log_store.add_step_log(step_log, self._context.run_id)


class MapNode(CompositeNode):
    """
    A composite node that contains ONE graph object within itself that has to be executed with an iterable.

    The structure is generally:
        MapNode:
            branch

        The config is expected to have a variable 'iterate_on' and iterate_as which are looked for in the parameters.
        for iter_variable in parameters['iterate_on']:
            Execute the Branch by sending {'iterate_as': iter_variable}

    The internal naming convention creates branches dynamically based on the iteration value
    """

    node_type: str = Field(default="map", serialization_alias="type")
    iterate_on: str
    iterate_as: str
    branch: Graph
    is_composite: bool = True

    @classmethod
    def parse_from_config(cls, config: Dict[str, Any]) -> "MapNode":
        internal_name = cast(str, config.get("internal_name"))

        config_branch = config.pop("branch", {})
        if not config_branch:
            raise Exception("A map node should have a branch")

        branch = create_graph(
            deepcopy(config_branch),
            internal_branch_name=internal_name + "." + defaults.MAP_PLACEHOLDER,
        )
        return cls(branch=branch, **config)

    def _get_branch_by_name(self, branch_name: str) -> Graph:
        """
        Retrieve a branch by name.

        In the case of a Map Object, the branch naming is dynamic as it is parameterized on iterable.
        This method takes no responsibility in checking the validity of the naming.

        Returns a Graph Object

        Args:
            branch_name (str): The name of the branch to retrieve

        Raises:
            Exception: If the branch by that name does not exist
        """
        return self.branch

    def fan_out(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        The general method to fan out for a node of type map.
        This method assumes that the step log has already been created.

        3rd party orchestrators should call this method to create the individual branch logs.

        Args:
            executor (BaseExecutor): The executor class as defined by the config
            map_variable (dict, optional): If the node is part of map. Defaults to None.
        """
        iterate_on = self._context.run_log_store.get_parameters(self._context.run_id)[self.iterate_on]

        # Prepare the branch logs
        for iter_variable in iterate_on:
            effective_branch_name = self._resolve_map_placeholders(
                self.internal_name + "." + str(iter_variable), map_variable=map_variable
            )
            branch_log = self._context.run_log_store.create_branch_log(effective_branch_name)
            branch_log.status = defaults.PROCESSING
            self._context.run_log_store.add_branch_log(branch_log, self._context.run_id)

    def execute_as_graph(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        This function does the actual execution of the branch of the map node.

        From a design perspective, this function should not be called if the execution is 3rd party orchestrated.

        The modes that render the job specifications, do not need to interact with this node at all as
        they have their own internal mechanisms of handing map states or dynamic parallel states.
        If they do not, you can find a way using as-is nodes as hack nodes.

        The actual logic is :
            * We iterate over the iterable as mentioned in the config
            * For every value in the iterable we call the executor.execute_graph(branch, iterate_as: iter_variable)

        The execution of a dag, could result in
            * The dag being completely executed with a definite (fail, success) state in case of local
                or local-container execution
            * The dag being in a processing state with PROCESSING status in case of local-aws-batch

        Only fail state is considered failure during this phase of execution.

        Args:
            executor (Executor): The Executor as per the use config
            map_variable (dict): The map variables the graph belongs to
            **kwargs: Optional kwargs passed around
        """
        from magnus import entrypoints

        iterate_on = None
        try:
            iterate_on = self._context.run_log_store.get_parameters(self._context.run_id)[self.iterate_on]
        except KeyError:
            raise Exception(
                f"Expected parameter {self.iterate_on} not present in Run Log parameters, was it ever set before?"
            )

        if not isinstance(iterate_on, list):
            raise Exception("Only list is allowed as a valid iterator type")

        self.fan_out(map_variable=map_variable, **kwargs)

        jobs = []
        # Given that we can have nesting and complex graphs, controlling the number of processess is hard.
        # A better way is to actually submit the job to some process scheduler which does resource management
        for iter_variable in iterate_on:
            effective_map_variable = map_variable or OrderedDict()
            effective_map_variable[self.iterate_as] = iter_variable

            if self._context.executor._is_parallel_execution():
                # Trigger parallel jobs
                action = entrypoints.execute_single_brach
                kwargs = {
                    "configuration_file": self._context.configuration_file,
                    "pipeline_file": self._context.pipeline_file,
                    "branch_name": self.branch.internal_branch_name.replace(" ", defaults.COMMAND_FRIENDLY_CHARACTER),
                    "run_id": self._context.run_id,
                    "map_variable": json.dumps(effective_map_variable),
                    "tag": self._context.tag,
                }
                process = multiprocessing.Process(target=action, kwargs=kwargs)
                jobs.append(process)
                process.start()

            else:
                # If parallel is not enabled, execute them sequentially
                self._context.executor.execute_graph(self.branch, map_variable=effective_map_variable, **kwargs)

        for job in jobs:
            job.join()

        self.fan_in(map_variable=map_variable, **kwargs)

    def fan_in(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        The general method to fan in for a node of type map.

        3rd  party orchestrators should call this method to find the status of the step log.

        Args:
            executor (BaseExecutor): The executor class as defined by the config
            map_variable (dict, optional): If the node is part of map node. Defaults to None.
        """
        iterate_on = self._context.run_log_store.get_parameters(self._context.run_id)[self.iterate_on]
        # # Find status of the branches
        step_success_bool = True

        for iter_variable in iterate_on:
            effective_branch_name = self._resolve_map_placeholders(
                self.internal_name + "." + str(iter_variable), map_variable=map_variable
            )
            branch_log = self._context.run_log_store.get_branch_log(effective_branch_name, self._context.run_id)
            if branch_log.status != defaults.SUCCESS:
                step_success_bool = False

        # Collate all the results and update the status of the step
        effective_internal_name = self._resolve_map_placeholders(self.internal_name, map_variable=map_variable)
        step_log = self._context.run_log_store.get_step_log(effective_internal_name, self._context.run_id)

        if step_success_bool:  #  If none failed and nothing is waiting
            step_log.status = defaults.SUCCESS
        else:
            step_log.status = defaults.FAIL

        self._context.run_log_store.add_step_log(step_log, self._context.run_id)


class DagNode(CompositeNode):
    """
    A composite node that internally holds a dag.

    The structure is generally:
        DagNode:
            dag_definition: A YAML file that holds the dag in 'dag' block

        The config is expected to have a variable 'dag_definition'.
    """

    node_type: str = Field(default="dag", serialization_alias="type")
    dag_definition: str
    branch: Graph
    is_composite: bool = True
    internal_branch_name: Annotated[str, Field(validate_default=True)] = ""

    @field_validator("internal_branch_name")
    @classmethod
    def validate_internal_branch_name(cls, internal_branch_name: str, info: ValidationInfo):
        internal_name = info.data["internal_name"]
        return internal_name + "." + defaults.DAG_BRANCH_NAME

    @field_validator("dag_definition")
    @classmethod
    def validate_dag_definition(cls, value):
        if not value.endswith(".yaml"):  # TODO: Might have a problem with the SDK
            raise ValueError("dag_definition must be a YAML file")
        return value

    @classmethod
    def parse_from_config(cls, config: Dict[str, Any]) -> "DagNode":
        internal_name = cast(str, config.get("internal_name"))

        if "dag_definition" not in config:
            raise Exception(f"No dag definition found in {config}")

        dag_config = utils.load_yaml(config["dag_definition"])
        if "dag" not in dag_config:
            raise Exception("No DAG found in dag_definition, please provide it in dag block")

        branch = create_graph(dag_config["dag"], internal_branch_name=internal_name + "." + defaults.DAG_BRANCH_NAME)

        return cls(branch=branch, **config)

    def _get_branch_by_name(self, branch_name: str):
        """
        Retrieve a branch by name.
        The name is expected to follow a dot path convention.

        Returns a Graph Object

        Args:
            branch_name (str): The name of the branch to retrieve

        Raises:
            Exception: If the branch_name is not 'dag'
        """
        if branch_name != self.internal_branch_name:
            raise Exception(f"Node of type {self.node_type} only allows a branch of name {defaults.DAG_BRANCH_NAME}")

        return self.branch

    def fan_out(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        The general method to fan out for a node of type dag.
        The method assumes that the step log has already been created.

        Args:
            executor (BaseExecutor): The executor class as defined by the config
            map_variable (dict, optional): _description_. Defaults to None.
        """
        effective_branch_name = self._resolve_map_placeholders(self.internal_branch_name, map_variable=map_variable)

        branch_log = self._context.run_log_store.create_branch_log(effective_branch_name)
        branch_log.status = defaults.PROCESSING
        self._context.run_log_store.add_branch_log(branch_log, self._context.run_id)

    def execute_as_graph(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        This function does the actual execution of the branch of the dag node.

        From a design perspective, this function should not be called if the execution is 3rd party orchestrated.

        The modes that render the job specifications, do not need to interact with this node at all
        as they have their own internal mechanisms of handling sub dags.
        If they do not, you can find a way using as-is nodes as hack nodes.

        The actual logic is :
            * We just execute the branch as with any other composite nodes
            * The branch name is called 'dag'

        The execution of a dag, could result in
            * The dag being completely executed with a definite (fail, success) state in case of
                local or local-container execution
            * The dag being in a processing state with PROCESSING status in case of local-aws-batch

        Only fail state is considered failure during this phase of execution.

        Args:
            executor (Executor): The Executor as per the use config
            **kwargs: Optional kwargs passed around
        """
        self.fan_out(map_variable=map_variable, **kwargs)
        self._context.executor.execute_graph(self.branch, map_variable=map_variable, **kwargs)
        self.fan_in(map_variable=map_variable, **kwargs)

    def fan_in(self, map_variable: TypeMapVariable = None, **kwargs):
        """
        The general method to fan in for a node of type dag.

        3rd party orchestrators should call this method to find the status of the step log.

        Args:
            executor (BaseExecutor): The executor class as defined by the config
            map_variable (dict, optional): If the node is part of type dag. Defaults to None.
        """
        step_success_bool = True
        effective_branch_name = self._resolve_map_placeholders(self.internal_branch_name, map_variable=map_variable)
        effective_internal_name = self._resolve_map_placeholders(self.internal_name, map_variable=map_variable)

        branch_log = self._context.run_log_store.get_branch_log(effective_branch_name, self._context.run_id)
        if branch_log.status != defaults.SUCCESS:
            step_success_bool = False

        step_log = self._context.run_log_store.get_step_log(effective_internal_name, self._context.run_id)
        step_log.status = defaults.PROCESSING

        if step_success_bool:  #  If none failed and nothing is waiting
            step_log.status = defaults.SUCCESS
        else:
            step_log.status = defaults.FAIL

        self._context.run_log_store.add_step_log(step_log, self._context.run_id)


class StubNode(ExecutableNode):
    """
    Stub is a convenience design node.

    It always returns success in the attempt log and does nothing.

    This node is very similar to pass state in Step functions.

    This node type could be handy when designing the pipeline and stubbing functions
    """

    node_type: str = Field(default="stub", serialization_alias="type")
    model_config = ConfigDict(extra="allow")

    @classmethod
    def parse_from_config(cls, config: Dict[str, Any]) -> "StubNode":
        return cls(**config)

    def execute(self, mock=False, map_variable: TypeMapVariable = None, **kwargs) -> StepAttempt:
        """
        Do Nothing node.
        We just send an success attempt log back to the caller

        Args:
            executor ([type]): [description]
            mock (bool, optional): [description]. Defaults to False.
            map_variable (str, optional): [description]. Defaults to ''.

        Returns:
            [type]: [description]
        """
        attempt_log = self._context.run_log_store.create_attempt_log()

        attempt_log.start_time = str(datetime.now())
        attempt_log.status = defaults.SUCCESS  # This is a dummy node and always will be success

        attempt_log.end_time = str(datetime.now())
        attempt_log.duration = utils.get_duration_between_datetime_strings(attempt_log.start_time, attempt_log.end_time)
        return attempt_log
